#ifndef _FLIB_H_
#define _FLIB_H_

#include <stdlib.h>
#include <assert.h>
#include "flib_internal.h"
#include "flib_tests.h"
#include "flib_groups.h"
#include "flib_xor.h"
#include "libtopodisc.h"
#include <limits.h>
#include "flib_debug.h"

#define _FLIB_NONE_HIERARCHY  888
#define _FLIB_NODES_HIERARCHY 889

/* control tags */
#define CP_DEMAND_PUT_CHECKPOINT_REQUEST							121
#define CP_DEMAND_PUT_CHECKPOINT_DONE									122
#define CP_DEMAND_GET_CHECKPOINT_REQUEST							123
#define CP_DEMAND_GET_CHECKPOINT_DONE									124
#define CP_INITIAL_CHECKPOINT_DONE										125
#define CP_MASTER_COORDINATED_CHECKPOINT_REQUEST			126
#define CP_COORDINATED_CHECKPOINT_REQUEST							127
#define CP_RECOVERY_CHECKPOINT												128
#define CP_NECESSARY_CHECKPOINT_DONE									129	
#define CP_MASTER_GLOBAL_RECOVERY_REQUEST							130
#define CP_GLOBAL_RECOVERY_REQUEST										131
#define CP_GLOBAL_LOGS_CLEARED												132
#define __CM_GLOBAL_LOGS_CLEARED											133

#define CM_DEMAND_PUT_CHECKPOINT_REQUEST		454
#define CM_DEMAND_PUT_CHECKPOINT_DONE				455
#define CM_DEMAND_GET_CHECKPOINT_REQUEST		456
#define CM_DEMAND_GET_CHECKPOINT_DONE				457
#define CM_COORDINATED_CHECKPOINT						458
#define CM_PROC_ERROR												459
#define CM_PROC_LOCAL_RECOVERY							460
#define CM_PROC_GLOBAL_RECOVERY							555
#define CM_INITIAL_CHECKPOINT								461
#define CM_QUIT															462
#define CM_FAKED_COORDINATED_CHECKPOINT			463
#define CM_NECESSARY_CHECKPOINT_DONE				464
#define CM_GLOBAL_RECOVERY_CONFIRMATION			465
#define CM_GLOBAL_LOGS_CLEARED							466

#define ADAPTATION_FREQUENCY	10101

#define FLIB_APP_NONBLCK	771166
#define FLIB_APP_BLK			771165

/* debug tags */
#define DEBUG_CM_INITIAL_CONFIRMATION		900
#define DEBUG_CP_CHECKPOINT_TEST				901

#define FLIB_NO_BSP                 90                                                                                                                                        
#define FLIB_BSP_DISTRIBUTED    		91
#define FLIB_BSP_SIMPLE             92                                                                                                                                            
#define FLIB_NO_COORDINATED_SCHEME  93
#define FLIB_ORDERED_CHECKPOINT			94

#define FLIB_AVERAGE	9999
#define FLIB_MAX			9998
#define FLIB_ALL			9997

#define FLIB_LOGGING_GETS_LOCK		888888
#define FLIB_LOGGING_GETS_ATOMICS	888887

#define TRUE 1
#define FALSE 0

#ifdef __cplusplus
 extern "C" {
#endif

/* general protocol user functions */

int fMPI_Win_create(void *base, MPI_Aint size, int disp_unit, MPI_Info info, MPI_Comm comm, foMPI_Win *win, int cp_procs, MPI_Comm* user_comm, int hierarchy, double mtbf, int protocol_type, int log_puts, int log_gets, int log_list_size);
int fMPI_Win_allocate(MPI_Aint size, int disp_unit, MPI_Info info, MPI_Comm comm, void *baseptr, foMPI_Win *win, int cp_procs, MPI_Comm* user_comm, int hierarchy, double mtbf, int protocol_type, int log_puts, int log_gets, int log_list_size);
int fMPI_Win_free(foMPI_Win* win);

/* useful functions */

int fLIB_get_user_rank();
int fLIB_get_user_size();
double fLIB_get_daly_interval();

int is_cp_proc(int initial_user_rank);
int is_cm_proc(int initial_user_rank);
int cp_proc();
int cm_proc();

void recover_me();
void fLIB_take_ordered_coordinated_checkpoint();
void fLIB_local_recovery();
void fLIB_global_recovery();

//TODO: only for the sake of simulation, remove in final version!
void _fLIB_handle_cp_global_recovery_request();
void _fLIB_cm_manage_global_recovery();
void __test_all_logs_of_gets(int verbose);
void __replay_actions();

void clear_all_counters();
void clear_all_logs();

//Functions that should be used instead of original ones

int fMPI_Win_fence(int assert, foMPI_Win win);

int fMPI_Win_lock(int lock_type, int rank, int assert, foMPI_Win win);
int fMPI_Win_unlock(int rank, foMPI_Win win);
int fMPI_Win_lock_all(int assert, foMPI_Win win);
int fMPI_Win_unlock_all(foMPI_Win win);

int fMPI_Win_flush(int rank, foMPI_Win win);
int fMPI_Win_flush_all(foMPI_Win win);
int fMPI_Win_flush_local(int rank, foMPI_Win win);
int fMPI_Win_flush_local_all(foMPI_Win win);
int fMPI_Win_sync(foMPI_Win win);

int fMPI_Put(const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype, foMPI_Win win);
int fMPI_Get(void *origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype, foMPI_Win win);
int fMPI_Accumulate(const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count,
                     MPI_Datatype target_datatype, MPI_Op op, foMPI_Win win);
int fMPI_Get_accumulate(const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, void *result_addr, int result_count, MPI_Datatype result_datatype,
                         int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype, MPI_Op op, foMPI_Win win);

int fMPI_Fetch_and_op(const void *origin_addr, void *result_addr, MPI_Datatype datatype, int target_rank, MPI_Aint target_disp, MPI_Op op, foMPI_Win win);
int fMPI_Compare_and_swap(const void *origin_addr, const void *compare_addr, void *result_addr, MPI_Datatype datatype, int target_rank, MPI_Aint target_disp, foMPI_Win win);

void debug_message(int proc_to_speak, int fail, FILE* stream, const char* format, ...);
void debug_message_e(int proc_to_speak, int fail, FILE* stream, const char* format);

void debug_message_gen(int fail, FILE* stream, const char* format, ...);
void debug_message_gen_e(int fail, FILE* stream, const char* format);

#ifdef __cplusplus
  }
#endif

#endif
