#include "flib_cm_proc.h"
#include "flib_tests.h"
#include "flib_debug.h"
#include "flib_logging_protocol.h"
#include "flib_recovery.h"

//#######################################################################################################################################
//main notification loop
//#######################################################################################################################################

void cm_process_notifications() {
#ifdef MEASURE_LOGGING_OVERHEADS
	return;
#endif
	assert(cm_proc());

	//TODO: check if analyzing get logging space here helps with demand checkpoints
	analyze_get_logging_space(0);

	//only if the application is nonblocking (no fences etc) we can check it at every call
	//otherwise, we check if after every barrier/fence
	if((*desc)->app_type == FLIB_APP_NONBLCK) {
		cm_manage_global_recovery();
	}

	MPI_Status s;
	MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &((*desc)->awaiting_requests), &s);	

	while((*desc)->awaiting_requests == TRUE) {
		(*desc)->awaiting_requests = FALSE;

		int sender = s.MPI_SOURCE;
		int tag = s.MPI_TAG;
		if(tag != __CM_GLOBAL_LOGS_CLEARED) {assert(is_cp_proc(sender));}

		if(tag == CP_DEMAND_PUT_CHECKPOINT_REQUEST) {
			handle_cp_demand_put_checkpoint_request();
		}
		else if(tag == CP_DEMAND_PUT_CHECKPOINT_DONE) {
			handle_cp_demand_put_checkpoint_done();
		}
		else if(tag == CP_DEMAND_GET_CHECKPOINT_REQUEST) {
			handle_cp_demand_get_checkpoint_request();
		}
		else if(tag == CP_DEMAND_GET_CHECKPOINT_DONE) {
			handle_cp_demand_get_checkpoint_done();
		}
		else if(tag == __CM_GLOBAL_LOGS_CLEARED) {
			__handle_cm_global_logs_cleared();
		}
		else if(tag == CP_NECESSARY_CHECKPOINT_DONE) {
			handle_cp_necessary_checkpoint_done();
		}
		else if(tag == CP_COORDINATED_CHECKPOINT_REQUEST) {
			start_cp_coordinated_checkpoint_request();
		}
		else if(tag == CP_GLOBAL_RECOVERY_REQUEST) {
			handle_cp_global_recovery_request();
		}
		else {
			handle_unknown_request_cm(tag);	
		}

		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &((*desc)->awaiting_requests), &s);
	}	
}

//#######################################################################################################################################
//PUT demand checkpoint requests
//#######################################################################################################################################

void handle_cp_demand_put_checkpoint_request() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot receive cp demand checkpoint request if no logging is enabled!\n");
	}
	MPI_Status s;
	int origin_rank = -1;

	MPI_Recv(&origin_rank, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);
	assert(s.MPI_SOURCE == assigned_cp_proc((*desc)->initial_user_rank));

	origin_rank = translate_from_world_to_flib_rank(origin_rank);

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CM: Received demand PUT checkpoint request for [%i,IN] (from CP: [%i,IN])\n", origin_rank, s.MPI_SOURCE);
#endif

	//TODO: RMA accumulate! - finish streaming impl
	int* data = take_demand_checkpoint(origin_rank);
	//TODO: double check if we really don't need to send these epochs for demand PUT checkpoints
	//int data_to_send[3];
	//data_to_send[0] = data[0];
	//data_to_send[1] = data[1];
	//data_to_send[2] = origin_rank;

	MPI_Send(&origin_rank, 1, MPI_INT, s.MPI_SOURCE, CM_DEMAND_PUT_CHECKPOINT_DONE, (*desc)->initial_user_comm);
	free(data);

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CM: Demand PUT checkpoint done, sent confirmation to CP [%i,IN] (bask to CM [%i,IN])\n", s.MPI_SOURCE, origin_rank);
#endif
}

void handle_cp_demand_put_checkpoint_done() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot receive cp demand checkpoint done if no logging is enabled!\n");
	}
	MPI_Status s;
	//int data[3];
	int target_rank;
	MPI_Recv(&target_rank, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	(*cm_desc)->sent_cm_adaptive_requests--;
	assert((*cm_desc)->sent_cm_adaptive_requests >= 0);

	//int epoch_to_clear_local = data[0];
	//int epoch_to_clear_global = data[1];
	//int origin_rank = data[2];
	int epoch_to_clear_local = (*cm_desc)->sent_checkpoint_request_epoch[target_rank];
	assert(epoch_to_clear_local != _FLIB_DEMAND_PUT_NOT_SENT && epoch_to_clear_local >= 0);
	(*cm_desc)->sent_checkpoint_request_epoch[target_rank] = _FLIB_DEMAND_PUT_NOT_SENT;
	(*cm_desc)->sent_checkpoint_request[target_rank] = FALSE;

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr,
			"*** CM: Received demand PUT checkpoint done confirmation from CM [%i,IN] (from CP: [%i,IN]), still %i requests awaiting. Begin to clear the LOG...\n",
			target_rank, s.MPI_SOURCE, (*cm_desc)->sent_cm_adaptive_requests);
#endif

	clear_local_logging_space(target_rank, epoch_to_clear_local);

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr,
			"*** CM: Local logs cleared, remaining space (local): %i\n",
			(*desc)->flib_parameters->log_list_size - (*cm_desc)->log_space->next_free_local_offset);
#endif
}


//#######################################################################################################################################
//GET checkpoint requests
//#######################################################################################################################################

void handle_cp_demand_get_checkpoint_request() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot receive cp get demand checkpoint request if no logging is enabled!\n");
	}
	MPI_Status s;
	int origin_rank = -1;

	MPI_Recv(&origin_rank, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

#ifndef MEASURE_LOGGING_OVERHEADS
	assert(s.MPI_SOURCE == assigned_cp_proc((*desc)->initial_user_rank));
	assert(s.MPI_TAG == CP_DEMAND_GET_CHECKPOINT_REQUEST);
#endif

	int origin_rank_cm = translate_from_world_to_flib_rank(origin_rank);

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CM: Received demand GET checkpoint request for [%i,IN] (from CP: [%i,IN])\n", origin_rank, s.MPI_SOURCE);
#endif

	int* data = take_demand_checkpoint(origin_rank_cm);
	int data_to_send[2];
	data_to_send[0] = origin_rank;
	data_to_send[1] = data[0];

	MPI_Send(&data_to_send[0], 2, MPI_INT, s.MPI_SOURCE, CM_DEMAND_GET_CHECKPOINT_DONE, (*desc)->initial_user_comm);
	free(data);

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr, "*** CM: Demand GET checkpoint done, sent confirmation to CP [%i,IN] (bask to CM [%i,IN])\n", s.MPI_SOURCE, origin_rank);
#endif
}

void handle_cp_demand_get_checkpoint_done() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot receive cp get demand checkpoint done if no logging is enabled!\n");
	}
	MPI_Status s;
	int data[2];
	int target_rank;
	int epoch_to_clear;

	MPI_Recv(&data[0], 2, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);
	target_rank = data[0];
	epoch_to_clear = data[1];
	int rank_to_clear_cm = translate_from_world_to_flib_rank(target_rank);

	(*cm_desc)->sent_cm_get_demand_requests--;
	assert((*cm_desc)->sent_cm_get_demand_requests >= 0);

	(*cm_desc)->sent_checkpoint_request[target_rank] = FALSE;

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr,
			"*** CM: Received demand GET checkpoint done confirmation from CM [%i,IN] (from CP: [%i,IN]), still %i requests awaiting. Begin to clear the LOG...\n",
			target_rank, s.MPI_SOURCE, (*cm_desc)->sent_cm_get_demand_requests);
#endif

	clear_global_logging_space(rank_to_clear_cm, epoch_to_clear);

#ifdef FLIB_LOGGING_GETS_DEBUG
	(*cm_desc)->log_space->__cleared_global_logs--;

	MPI_Send(&epoch_to_clear, 1, MPI_INT, target_rank, __CM_GLOBAL_LOGS_CLEARED, (*desc)->initial_user_comm);
#endif
}

//#######################################################################################################################################
//cleared logs messages
//#######################################################################################################################################

void __handle_cm_global_logs_cleared() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### ERROR: Cannot receive __cm global logs cleared if no logging is enabled!\n");
	}
	MPI_Status s;
	int epoch_to_clear;
	MPI_Recv(&epoch_to_clear, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);	
	
	int rank_to_clear_cm = translate_from_world_to_flib_rank(s.MPI_SOURCE);

	__delete_get_testing_logs(rank_to_clear_cm, epoch_to_clear, TRUE);

	(*cm_desc)->log_space->__cleared_global_test_logs++;
}

//#######################################################################################################################################
//NECESSARY checkpoint messages
//#######################################################################################################################################

//TODO: merge these necessary checkpoints with demand checkpoints
void handle_cp_necessary_checkpoint_done() {
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot receive cp necessary checkpoint done if no logging is enabled!\n");
	}
	MPI_Status s;
	int data[2];
	MPI_Recv(&data[0], 2, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);	

	int epoch_to_clear = data[0];
	int origin_rank = data[1];
	//int world_origin_rank = translate_from_flib_to_world_rank(origin_rank);	TODO: double check

#ifdef FLIB_DEMAND_DEBUG
	debug_message_gen( FALSE, stderr,
			"*** CM: Received NECESSARY checkpoint done confirmation from CM [%i,IN] (from CP: [%i,IN]). Begin to clear the GET LOG...\n",
			origin_rank, s.MPI_SOURCE);
#endif

	//TODO: make sure that we want this
	//clear_global_logging_space(origin_rank, epoch_to_clear);
	
	analyze_get_logging_space(0);

#ifdef FLIB_DEMAND_DEBUG
	//debug_message_gen( FALSE, stderr,
	//		"*** CM: Global GET logs cleared, remaining space (local): %i\n",
	//		((double)(*desc)->flib_parameters->log_list_size - (double)get_remaining_global_space())/(double)(*desc)->flib_parameters->log_list_size);
#endif

	MPI_Send(&origin_rank, 1, MPI_INT, s.MPI_SOURCE, CM_GLOBAL_LOGS_CLEARED, (*desc)->initial_user_comm);
}

//#######################################################################################################################################
//demand checkpoint routines
//#######################################################################################################################################

// this checkpoint is the demand checkpoint which only compute processes can take 
int* take_demand_checkpoint(int requesting_rank) {	
	if(!do_logging()) {
		debug_message_gen_e(TRUE, stderr,"########### Cannot take demand checkpoint if no logging enables!\n");
	}
	assert(cm_proc());

	uint64_t* app_data = (uint64_t*) ((*((*cm_desc)->app_win))->win_array)[(*desc)->flib_level_rank].base;
	uint64_t* epoch_data = (uint64_t*) (*cm_desc)->epoch_data;	

	uint32_t app_data_size = (*desc)->app_data_size; 
	uint32_t epoch_data_size = (*desc)->epoch_data_size; 

	uint64_t* checkpoint_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->app_data_offset];
	uint64_t* checkpoint_app_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->app_data_offset];
	uint64_t* checkpoint_epoch_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->epoch_data_offset];

	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->xor_data_offset];

	//TODO: check, and what if the user self-locked before?
	foMPI_Win_flush_all(*(*cm_desc)->app_win);
	foMPI_Win_flush_all(*(*cm_desc)->epoch_win);
	foMPI_Win_flush_all(*(*desc)->cp_win);

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(*cm_desc)->app_win);
	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(*cm_desc)->epoch_win);	
	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->initial_user_rank, 0, *(*desc)->cp_win);

	/* first, we copy the whole bulk of or old CP data to the xor part of the window */
	memcpy(xored_data, checkpoint_data, app_data_size + epoch_data_size);

	/* then, we copy the currect application data to the adequate place in the window */
	memcpy(checkpoint_app_data, app_data, app_data_size);

	/* then, we copy the current epoch counter to the adequate place in the window */
	memcpy(checkpoint_epoch_data, epoch_data, epoch_data_size);

	/* finally, we xor the old CP (app data + epoch counter) with the current data and store the result in the xor-part of the window */
	calculate_xor(checkpoint_data, xored_data, app_data_size + epoch_data_size);

	//only the first one matters now, TODO: make sure the other can be deleted
	int* epoch_return_data = malloc( 2 * sizeof(int) );

	epoch_return_data[0] = (*cm_desc)->local_epoch_counter[requesting_rank];
	epoch_return_data[1] = (*cm_desc)->epoch_data[requesting_rank];

	foMPI_Win_unlock((*desc)->initial_user_rank, *(*desc)->cp_win);
	foMPI_Win_unlock((*desc)->flib_level_rank, *(*cm_desc)->epoch_win);
	foMPI_Win_unlock((*desc)->flib_level_rank, *(*cm_desc)->app_win);

#ifdef FLIB_VERBOSE_DEBUG
	debug_message_gen_e( FALSE, stderr, "*** CM: Took a demand checkpoint\n");
#endif

	return epoch_return_data;
}



//#######################################################################################################################################
//A recovery scheme (confirm to our CP that we enter the 'recovery phase')
//#######################################################################################################################################

void handle_cp_global_recovery_request() {
	assert(cm_proc());

	int assigned_proc = my_assigned_cp_proc();
	MPI_Status s;
	MPI_Recv(NULL, 0, MPI_CHAR, assigned_proc, CP_GLOBAL_RECOVERY_REQUEST, (*desc)->initial_user_comm, &s);

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_gen( FALSE, stderr, "^^^ CM: received global recovery request from CP: %i\n", s.MPI_SOURCE);
#endif

	MPI_Send(NULL, 0, MPI_CHAR, assigned_proc, CM_GLOBAL_RECOVERY_CONFIRMATION, (*desc)->initial_user_comm);

	NBC_Ibarrier(*(*desc)->flib_level_comm, &((*cm_desc)->global_recovery_request) );
	(*desc)->global_recovery_started = TRUE;
}

//#######################################################################################################################################
//A recovery scheme (finish the recovery in response to a request by its CP)
//#######################################################################################################################################

void cm_manage_global_recovery() {
	if((*desc)->global_recovery_started == FALSE) {
		return;
	}

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_gen_e( FALSE, stderr, "^^^ beginning to manage global recovery...\n");
#endif

	int finished = NBC_CONTINUE;
	MPI_Status s;

	//TODO: make this whole scheme asynchronous
	//while(finished == NBC_CONTINUE) NBC_Test( &((*cm_desc)->global_recovery_request), &finished, &s);
	//if(finished == NBC_OK) {
#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
		debug_message_gen_e( FALSE, stderr, "^^^ ibarrier done, begin to finish recovery...\n");
#endif
		uint64_t* checkpoint_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->data_offset];
		uint64_t* app_data = (uint64_t*) ((*((*cm_desc)->app_win))->win_array)[(*desc)->flib_level_rank].base;

		int app_data_size = (*desc)->app_data_size;
		int cp_data_unit = (*desc)->cp_data_disp_unit;

		memcpy(app_data, checkpoint_data, app_data_size);

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
		debug_message_gen_e( FALSE, stderr, "^^^ data copied...\n");
#endif

		//TODO: double check if it causes no deadlock
		finish_global_recovery();

		(*desc)->global_recovery_started = FALSE;
	//}
}

//#######################################################################################################################################
//CC routines
//#######################################################################################################################################

void start_cp_coordinated_checkpoint_request() {
	assert(cm_proc());
	assert((*desc)->coordinated_protocol_type == FLIB_BSP_DISTRIBUTED || (*desc)->coordinated_protocol_type == FLIB_NO_BSP);

	int assigned_proc = my_assigned_cp_proc();
	MPI_Status s;
	MPI_Recv(NULL, 0, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

	assert(s.MPI_SOURCE == assigned_proc);
	assert(s.MPI_TAG == CP_COORDINATED_CHECKPOINT_REQUEST);

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen( FALSE, stderr, "&&& CM: received coordinated request from CP: %i\n", s.MPI_SOURCE);
#endif

	if((*desc)->coordinated_protocol_type == FLIB_BSP_DISTRIBUTED) {
		cm_begin_BSP_distributed_coordinated_checkpoint();
	}
	else if((*desc)->coordinated_protocol_type == FLIB_NO_BSP) {
		int faked = FALSE;
		finalize_coordinated_checkpoint(faked);
	}
}

void cm_begin_BSP_distributed_coordinated_checkpoint() {
	NBC_Ibarrier(*(*desc)->flib_level_comm, &((*cm_desc)->coordinated_checkpoint_request) );
	(*desc)->coordinated_checkpoint_started = TRUE;
}

void cm_manage_coordinated_checkpoint() {
	assert(cm_proc());
	if((*desc)->coordinated_protocol_type == FLIB_NO_COORDINATED_SCHEME || (*desc)->coordinated_protocol_type == FLIB_ORDERED_CHECKPOINT) {
		return;
	}

	//TODO: double check if this relocation is fine 
	//(*cm_desc)->superstep++;

#ifdef FLIB_COORDINATED_DEBUG
	debug_message( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "&&& Begin Superstep: %i\n", (*cm_desc)->superstep);
#endif

	if((*desc)->coordinated_protocol_type == FLIB_BSP_SIMPLE) {
		cm_manage_BSP_simple_coordinated_checkpoint();
	}
	else if((*desc)->coordinated_protocol_type == FLIB_BSP_DISTRIBUTED) {
		cm_manage_BSP_distributed_coordinated_checkpoint();
	}
	cm_manage_degeneration();
}

//manages the initiation of checkpoints started by a CM proc; takes place after a win fence call 
void cm_manage_BSP_simple_coordinated_checkpoint() {
	if((*cm_desc)->initial_coordinated_measurement == TRUE) {
		cm_measure_frequency_of_coordinated_checkpoints();
	}

	if((*cm_desc)->initial_coordinated_measurement == FALSE) {
		if((*cm_desc)->superstep % (*cm_desc)->checkpoint_frequency == 0) {
#ifdef FLIB_COORDINATED_DEBUG
			debug_message( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "&&& Superstep: %i. Beginning to take the checkpoint...\n", (*cm_desc)->superstep);
#endif
			assert((*desc)->coordinated_checkpoint_started != TRUE);
			(*desc)->coordinated_checkpoint_started = TRUE;

			int faked = FALSE;
			finalize_coordinated_checkpoint(faked);

			if((*cm_desc)->recompute_coordinated_checkpoint_frequency == TRUE) {
				cm_manage_recomputing_coordinated_checkpoint_frequency();
			}
			(*desc)->coordinated_checkpoint_started = FALSE;
			(*desc)->prev_cp_time = MPI_Wtime();
		}
	}
}

void cm_manage_recomputing_coordinated_checkpoint_frequency() {
	int recompute = 0;
	if(is_first_cm_process()) {
		double current_time = MPI_Wtime();
		double interval = current_time - (*desc)->prev_cp_time;
		(*cm_desc)->nr_of_intervals_for_avg++;
		(*cm_desc)->avg_interval = (((*cm_desc)->avg_interval * ((*cm_desc)->nr_of_intervals_for_avg - 1)) + interval) / (*cm_desc)->nr_of_intervals_for_avg;
		if((*cm_desc)->nr_of_intervals_for_avg > (*cm_desc)->initial_intervals_for_skipping) {	//check if we have to recompute the frequency
			if(abs(interval - (*cm_desc)->avg_interval) > (*cm_desc)->frequency_interval_threshold) {
				(*cm_desc)->nr_of_irregular_intervals++;
			}
			else {
				(*cm_desc)->nr_of_irregular_intervals = 0;
			}

#ifdef FLIB_COORDINATED_DEBUG
			debug_message( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "&&& RECOMPUTE DATA: interval: %f, nr_of_intervals: %i, avg_interval: %f, thresh: %f, irr_ints: %i...\n",
					interval, (*cm_desc)->nr_of_intervals_for_avg, (*cm_desc)->avg_interval, (*cm_desc)->frequency_interval_threshold, (*cm_desc)->nr_of_irregular_intervals);
#endif

			if((*cm_desc)->nr_of_irregular_intervals > (*cm_desc)->nr_interval_threshold) {
				recompute = 1;
#ifdef FLIB_COORDINATED_DEBUG
				debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "&&& Begin to RECOMPUTE the freqency interval...\n");
#endif
			}
		}
	}

	MPI_Bcast(&recompute, 1, MPI_INT, get_first_cm_process_flib_rank(), *(*desc)->flib_level_comm);

	if(recompute == 1) {
		(*cm_desc)->initial_coordinated_measurement == TRUE;
		(*desc)->initial_time = MPI_Wtime();
		(*cm_desc)->superstep = 0;
	}
}

void cm_measure_frequency_of_coordinated_checkpoints() {
	int checkpoint_frequency = -1;
	if(is_first_cm_process()) {
		double current_time = MPI_Wtime();
		double elapsed_time = current_time - (*desc)->initial_time;

#ifdef FLIB_COORDINATED_DEBUG
		debug_message_gen( FALSE, stderr, "&&& CM MASTER: Measurement. Elapsed time: %f. Superstep: %i\n", elapsed_time, (*cm_desc)->superstep);
#endif	

		if(elapsed_time > (*desc)->daly_interval) {
			(*cm_desc)->initial_coordinated_measurement = FALSE;
			checkpoint_frequency = (*cm_desc)->superstep;
			(*cm_desc)->frequency_interval_threshold = elapsed_time / 10.0;

			(*cm_desc)->nr_of_irregular_intervals = 0;
			(*cm_desc)->nr_of_intervals_for_avg = 0;
			(*cm_desc)->avg_interval = 0;

			(*desc)->prev_cp_time = MPI_Wtime();

#ifdef FLIB_COORDINATED_DEBUG
			debug_message_gen(FALSE, stderr, "&&& CM MASTER: Measurement FINISHED. Elapsed time: %f. Superstep: %i. Frequency: %i\n", 
					elapsed_time, (*cm_desc)->superstep, checkpoint_frequency);
#endif
		}
	}

	MPI_Bcast(&checkpoint_frequency, 1, MPI_INT, get_first_cm_process_flib_rank(), *(*desc)->flib_level_comm);

	if(checkpoint_frequency >= 0) {
#ifdef FLIB_COORDINATED_DEBUG
		debug_message( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "&&& Measured optimum frequency. The frequency is %i\n", checkpoint_frequency);
#endif
		(*cm_desc)->checkpoint_frequency = checkpoint_frequency;
		(*cm_desc)->initial_coordinated_measurement = FALSE;
	}
}

void cm_manage_BSP_distributed_coordinated_checkpoint() {
	int sum = 0;
	int result = 0;

	if((*desc)->coordinated_checkpoint_started == TRUE) {
		int finished = NBC_CONTINUE;
		MPI_Status s;
		NBC_Test( &((*cm_desc)->coordinated_checkpoint_request), &finished, &s);
		if(finished == NBC_OK) result = 1;
	}

	//TODO: this I comment out; double-check if the following faster is actually correct;
	//MPI_Allreduce(&result, &sum, 1, MPI_INT, MPI_SUM, *(*desc)->flib_level_comm);

	//TODO: is it correct??? (blocking)
	MPI_Reduce(&result, &sum, 1, MPI_INT, MPI_SUM, 0, *(*desc)->flib_level_comm);
	MPI_Bcast(&sum, 1, MPI_INT, 0, *(*desc)->flib_level_comm);

	if(sum == (*desc)->cm_size) {
#ifdef FLIB_COORDINATED_DEBUG
		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "&&& CM: begin to finalize the coordinated checkpoint\n");
#endif

		int faked = FALSE;
		finalize_coordinated_checkpoint(faked);

		(*desc)->coordinated_checkpoint_started = FALSE;
	}
	else {
#ifdef FLIB_COORDINATED_DEBUG
		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "&&& CM: distributed checkpoint still to be finalized\n");
#endif
	}
}

void finalize_coordinated_checkpoint(int faked) {
	flush_all_communication_channels();
	take_coordinated_checkpoint(faked);

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen_e( FALSE, stderr, "&&& Coordinated checkpoint taken\n");
#endif

	//TODO: double-check if necessary
	//MPI_Barrier((*desc)->initial_user_comm);
	MPI_Barrier(*(*desc)->flib_level_comm);

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen_e( FALSE, stderr, "&&& Final coordination barrier passed\n");
#endif
}

void flush_all_communication_channels() {
	foMPI_Win_fence(0, *(*cm_desc)->app_win);

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen_e( FALSE, stderr, "&&& FLushed all comm channels \n");
#endif
}

void take_coordinated_checkpoint(int faked) {
	/* TODO: based on Robert's implementation; make generic? */
	uint64_t* app_data = (uint64_t*) ((*((*cm_desc)->app_win))->win_array)[(*desc)->flib_level_rank].base;
	uint64_t* checkpoint_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->app_data_offset];
	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->xor_data_offset];
	uint32_t app_data_size = (*desc)->app_data_size;

	int assigned_proc = assigned_cp_proc((*desc)->initial_user_rank);
	int cp_data_unit = (*desc)->cp_data_disp_unit;

	//not using any locks because this function is always called inside two barriers/fences
	//TODO: double check all special corner-cases

	MPI_Request r;
	if(faked == FALSE) {
		MPI_Isend(app_data, app_data_size / cp_data_unit, MPI_UINT64_T, assigned_proc, CM_COORDINATED_CHECKPOINT, (*desc)->initial_user_comm, &r);
	}
	else {
		MPI_Isend(app_data, app_data_size / cp_data_unit, MPI_UINT64_T, assigned_proc, CM_FAKED_COORDINATED_CHECKPOINT, (*desc)->initial_user_comm, &r);
	}

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen( FALSE, stderr, "&&& sent COORDINATED checkpoint to %i \n", assigned_proc);
#endif

	/* in the coordinated (as in initial) checkpoint, we copy the application data to both parts of the checkpoint window */
	memcpy(checkpoint_data, app_data, app_data_size);
	memcpy(xored_data, app_data, app_data_size);

	/* we clear all the counters */
	memset((*cm_desc)->local_epoch_counter, 0, sizeof(int64_t) * (*desc)->cm_size);
	memset((*cm_desc)->epoch_data, 0, sizeof(int64_t) * (*desc)->cm_size);

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen_e( FALSE, stderr, "&&& memory copied and set \n");
#endif

	MPI_Status s;
	MPI_Wait(&r,&s);

#ifdef FLIB_COORDINATED_DEBUG
	debug_message_gen_e( FALSE, stderr, "&&& receipt confirmed \n");
#endif

	//TODO: check this!
	//free_all_outstanding_gets();

	clear_all_logs();
}

//#######################################################################################################################################
//Degeneration routines
//#######################################################################################################################################

void cm_manage_degeneration() {
	if(!do_logging()) {
		return;
	}
	assert(cm_proc());

	//TODO: leave it for now
	return;

	if((*cm_desc)->superstep <= 2 && (*desc)->degenerated == FALSE && (*desc)->coordinated_protocol_type != FLIB_NO_BSP) {
		/*TODO:make it not only for local*/
		int next_free = (*cm_desc)->log_space->next_free_local_offset;
		int remaining_space = (*desc)->flib_parameters->log_list_size - next_free;
		int spaces[(*desc)->cm_size];
		int remote_ops[(*desc)->cm_size];

		/*TODO: make this rank 0 abstract*/
		MPI_Gather(&remaining_space, 1, MPI_INT, &spaces[0], 1, MPI_INT, 0, *(*desc)->flib_level_comm);
		MPI_Gather(&((*desc)->remote_ops_done), 1, MPI_INT, &remote_ops[0], 1, MPI_INT, 0, *(*desc)->flib_level_comm);

#ifdef FLIB_DEBUG
		//fprintf(stderr,"[%i] gather performed\n", (*desc)->initial_user_rank);
#endif

		if(is_first_cm_process()) {
			int min_space_left = (*desc)->flib_parameters->log_list_size;
			int logging_list_size = min_space_left;			

			int i = 0;
			for(; i < (*desc)->cm_size; i++) {
				if(spaces[i] < min_space_left) {
					min_space_left = spaces[i];
				}
			}

			int min_ops = remote_ops[0];
			int ops_sum = 0;
			int communicated = 0;

			i = 0;
			for(; i < (*desc)->cm_size; i++) {
				ops_sum += remote_ops[i];
				if(remote_ops[i] < min_ops && remote_ops) {
					min_ops = remote_ops[i];
				}
				if(remote_ops[i] > 0) {
					communicated++;
				}
			}
			int ops_avg = ops_sum / (*desc)->cm_size;

			/*TODO: make this strategy better and based on user parameter*/
			if( ((double)min_space_left) / ((double)logging_list_size) <= 0.9 || (double)min_ops >= (double)(0.5*ops_avg) || (double)communicated >= (double)(0.3*(*desc)->cm_size)) {

#ifdef FLIB_DEBUG_DEGENERATION
				fprintf(stderr,"[%i] @@@@@@ undegenerated ---> degenerated\n", (*desc)->initial_user_rank);
#endif
				(*desc)->degenerated = TRUE;
			}

			MPI_Bcast(&((*desc)->degenerated), 1, MPI_INT, get_first_cm_process_flib_rank(), *(*desc)->flib_level_comm);
		}
		else {
			MPI_Bcast(&((*desc)->degenerated), 1, MPI_INT, get_first_cm_process_flib_rank(), *(*desc)->flib_level_comm);
		}

		if((*desc)->degenerated == TRUE) {	
#ifdef FLIB_DEBUG
			fprintf(stderr,"[%i] @@@ undegenerated ---> degenerated\n", (*desc)->initial_user_rank);
#endif
			clear_all_logs();
		}
	}
}

//#######################################################################################################################################
//other requests
//#######################################################################################################################################

void handle_unknown_request_cm(int tag) {
	debug_message_gen( TRUE, stderr, "########### CM: something went wrong, I received a message with an unknown tag: %i\n", tag);
}


