#include "flib_counters.h"
#include "flib_logging_protocol.h"

//#######################################################################################################################################
//Functions for managing various counters
//#######################################################################################################################################

//TODO: update to cover GNC, SC, and GC counters

void clear_all_counters() {
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	memset((*cm_desc)->local_epoch_counter, 0, sizeof(int64_t) * (*desc)->cm_size);

	//TODO: double check if this can be removed
	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *((*cm_desc)->epoch_win));
	memset((*cm_desc)->epoch_data, 0, sizeof(int64_t) * (*desc)->cm_size);
	foMPI_Win_unlock((*desc)->flib_level_rank, *((*cm_desc)->epoch_win));
}

void inc_global_epoch_counter(int rank) {
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	//TODO: double check if this is necessary
	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, rank, 0, *(*cm_desc)->epoch_win);

	int64_t add = 1;

	//foMPI_Put(&add, 1, MPI_INT, rank, (*desc)->flib_level_rank, 1, MPI_INT, *(*cm_desc)->epoch_win);
	//fprintf(stderr,"[%i] acc to %i, at disp: %i\n",(*desc)->initial_user_rank, rank, (*desc)->flib_level_rank);
	foMPI_Accumulate(&add, 1, MPI_INT64_T, rank, (*desc)->flib_level_rank, 1, MPI_INT64_T, foMPI_SUM, *(*cm_desc)->epoch_win);

	foMPI_Win_unlock(rank, *(*cm_desc)->epoch_win);
}

void inc_local_epoch_counter(int rank) {
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	(*cm_desc)->local_epoch_counter[rank]++;
}

void inc_local_epoch_counters() {
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	int i = 0;
	for(; i < (*desc)->cm_size; i++) {
		((*cm_desc)->local_epoch_counter)[i]++;
	}

#ifdef FLIB_EPOCH_COUNTER_DEBUG
	debug_message_gen( FALSE, stderr, "EEE All local epoch counters increased!!! counter at proc 0 is: %i\n", ((*cm_desc)->local_epoch_counter)[0]);
#endif
}

//TODO: double check if this is necessary
void inc_global_epoch_counters() {
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(*cm_desc)->epoch_win);

	int i = 0;
	for(; i < (*desc)->cm_size; i++) {
		((*cm_desc)->epoch_data)[i]++;
	}

	foMPI_Win_unlock((*desc)->flib_level_rank, *(*cm_desc)->epoch_win);
}

void inc_GNC_cnt() {
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	(*cm_desc)->GNC_cnt++;
}

void inc_GC_cnt() {
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	(*cm_desc)->GC_cnt++;
}

void inc_SC_cnt(int target_rank) {
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	//TODO:
}
