#include "flib.h"
#include "flib_groups.h"

int get_rank(MPI_Comm comm) {
	int rank = -1;
	MPI_Comm_rank(comm, &rank);
	return rank;
}

int get_initial_user_rank() {
	return (*desc)->initial_user_rank;
}

int get_size(MPI_Comm comm) {
	int size = -1;
	MPI_Comm_size(comm, &size);
	return size;
}

void print_comm_size(MPI_Comm* comm) {
	int size;
	MPI_Comm_size(*comm, &size);
	debug_message_gen( FALSE, stderr, ">>> COMM size: %i\n", size);
}

int get_initial_user_size() {
	return (*desc)->initial_user_size;
}

/* get initial user comm rank of the assigned CP proc */
int assigned_cp_proc(int cm_proc_rank) {
	if((*desc)->cp_size == 0) {
		//TODO
		debug_message_e( _FLIB_ANY_IN_INITIAL, TRUE, stderr, ">>> For now no CP procs is not supported; exiting...\n");
		return -1;
	}

	int rank = -1;

	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		rank = (int)(cm_proc_rank / (*desc)->group_size) * (*desc)->group_size;
	} 
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		rank = (int) cm_proc_rank % (*desc)->cp_size;
	}

	return rank;
}

int my_assigned_cp_proc() {	
	return assigned_cp_proc((*desc)->initial_user_rank);
}

int* get_cp_procs_ranks() {
	int* procs = malloc( (*desc)->cp_size * sizeof(int) );

	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		int i = 0;
		for(; i < (*desc)->cp_size; i++) {
			procs[i] = i * (*desc)->group_size;
		}
	} 
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		int i = 0;
		for(; i < (*desc)->cp_size; i++) {
			procs[i] = i;
		}
	}

	return procs;
}

// get world ranks of the assigned CM procs 
int* assigned_cm_procs() {
	assert(cp_proc());

	int* procs = (int*) malloc( sizeof(int) * assigned_cm_procs_nr() );
	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		int i = 0;
		for(; i < assigned_cm_procs_nr(); i++) {
			procs[i] = (*desc)->initial_user_rank + i + 1;
		}
	} 
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		int i = 1;
		for(; i <= assigned_cm_procs_nr(); i++) {
			procs[i-1] = (*desc)->initial_user_rank + i*(*desc)->cp_size;//i*(*desc)->node_level_size;
		}
	}

	return procs;	
}

int assigned_cm_procs_nr() {
	return (*desc)->group_size - 1;
}

int get_groups_nr(int cp_procs) {
	return cp_procs;
}

int get_group_size(int cp_procs) {
	//TODO
	if((*desc)->cp_size == 0) {
		return -1;
	}

	return get_initial_user_size() / get_groups_nr(cp_procs);
}

int group_size() {
	//TODO: address!!!!!!!
	//return get_initial_user_size() / (*desc)->cp_procs;
	assert(1 == 0);
	return -1;
}

int get_cm_procs_nr(int cp_procs) {
	return get_initial_user_size() - cp_procs;
}

void divide_into_groups(int hierarchy_scheme, int cp_procs, MPI_Comm initial_comm, MPI_Comm* flib_level_comm) {
	if(hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		divide_into_simple_scheme(cp_procs, initial_comm, flib_level_comm);
	}
	else if(hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		divide_into_nodes_scheme(cp_procs, initial_comm, flib_level_comm);
	}
}

void divide_into_simple_scheme(int cp_procs, MPI_Comm initial_comm, MPI_Comm* flib_level_comm) {
	MPI_Group initial_user_group;
	MPI_Group flib_level_group;

	MPI_Comm_group(initial_comm, &initial_user_group);

	(*desc)->initial_user_comm = initial_comm;

	int initial_user_rank = get_rank(initial_comm);
	int initial_user_size = get_size(initial_comm);

	(*desc)->initial_user_rank = initial_user_rank;
	(*desc)->initial_user_size = initial_user_size;

	(*desc)->cp_size = cp_procs;	
	(*desc)->cm_size = initial_user_size - cp_procs;

	(*desc)->groups_nr = cp_procs;

	if(cp_procs > 0) {
		(*desc)->group_size = initial_user_size / cp_procs;
	}
	else {
		//TODO: make it possible to not use CP's at all
		//useful for microbenchmarks
	}

	//our assumption about the nr of the groups
	assert(initial_user_size % (*desc)->groups_nr == 0);

	int ranks[1][3];
	ranks[0][0] = 0;
	ranks[0][1] = initial_user_size-1;
	ranks[0][2] = (*desc)->group_size;

	if(initial_user_rank % (*desc)->group_size == 0) {
		MPI_Group_range_incl(initial_user_group, 1, ranks, &flib_level_group);
		(*desc)->compute_proc = FALSE;
	}
	else {
		MPI_Group_range_excl(initial_user_group, 1, ranks, &flib_level_group);
		(*desc)->compute_proc = TRUE;
	}

	MPI_Comm_create(initial_comm, flib_level_group, flib_level_comm);
	(*desc)->flib_level_comm = flib_level_comm;
	(*desc)->flib_level_rank = get_rank(*(*desc)->flib_level_comm);

	//faked node rank, in this scheme this value is unnecessary
	(*desc)->node_level_rank = -1;
}


/*TODO: for now we assume that there are as many CP groups as the nodes*/
void divide_into_nodes_scheme(int cp_procs, MPI_Comm initial_comm, MPI_Comm* flib_level_comm) {
	// create a MPI communicator to identify onnode processes 
	int num;
	TPD_Get_num_hier_levels(&num, initial_comm);
	assert( num == 1 );

	TPD_Get_hier_levels(&((*desc)->node_level_comm), num, initial_comm);

	MPI_Comm_group( (*desc)->node_level_comm, &((*desc)->node_level_group) );
	(*desc)->node_level_rank = get_rank((*desc)->node_level_comm);

	int node_level_rank = get_rank((*desc)->node_level_comm);
	int node_level_size = get_size((*desc)->node_level_comm);

	(*desc)->node_level_rank = node_level_rank;
	(*desc)->node_level_size = node_level_size;

	//divide into FLIB comms

	MPI_Group initial_user_group;
	MPI_Group flib_level_group;

	MPI_Comm_group(initial_comm, &initial_user_group);

	(*desc)->initial_user_comm = initial_comm;

	int initial_user_rank = get_rank(initial_comm);
	int initial_user_size = get_size(initial_comm);

	(*desc)->initial_user_rank = initial_user_rank;
	(*desc)->initial_user_size = initial_user_size;

	(*desc)->cp_size = cp_procs;
	(*desc)->cm_size = initial_user_size - cp_procs;

	(*desc)->groups_nr = cp_procs;
	(*desc)->group_size = initial_user_size / cp_procs;

	//our assumption about the nr of the groups
	assert(initial_user_size % (*desc)->groups_nr == 0);

	//	if((*desc)->initial_user_rank == 0) {
	//		fprintf(stderr,"initial user size: %i, node level size: %i, cp size: %i, cm size: %i");

	//TODO: only for now
	assert((*desc)->cp_size % (*desc)->node_level_size == 0);

	int ranks[1][3];
	ranks[0][0] = 0;
	ranks[0][1] = (*desc)->cp_size-1;
	ranks[0][2] = 1;

	if(initial_user_rank < (*desc)->cp_size) {
		MPI_Group_range_incl(initial_user_group, 1, ranks, &flib_level_group);
		(*desc)->compute_proc = FALSE;
	}
	else {
		MPI_Group_range_excl(initial_user_group, 1, ranks, &flib_level_group);
		(*desc)->compute_proc = TRUE;
	}

	MPI_Comm_create(initial_comm, flib_level_group, flib_level_comm);
	(*desc)->flib_level_comm = flib_level_comm;
	(*desc)->flib_level_rank = get_rank(*(*desc)->flib_level_comm);

#ifdef FLIB_DEBUG
	//TODO: temporary, we should have independent number of CP groups and nodes
	__test_number_of_comms();
#endif
}

void __test_number_of_comms() {
#ifdef FLIB_DEBUG
	if((*desc)->compute_proc == TRUE) {
		fprintf(stderr,"[%i] CM PROCESS. flib rank: %i, node rank: %i\n",(*desc)->initial_user_rank, (*desc)->flib_level_rank, (*desc)->node_level_rank);
	}
	else {
		fprintf(stderr,"[%i] CP PROCESS. flib rank: %i, node rank: %i\n",(*desc)->initial_user_rank, (*desc)->flib_level_rank, (*desc)->node_level_rank);
	}
#endif

	//int assigned_ranks[(*desc)->group_size];

	//	MPI_Gather(&((*desc)->initial_user_rank), 1, MPI_INT, &assigned_ranks[0], 1, MPI_INT, 0, (*desc)->node_level_comm);

	if(cp_proc()) {
		int* assigned_procs = assigned_cm_procs();
		char procs_str[1024];
		procs_str[0] = '\0';

		int i = 0;
		for(; i < assigned_cm_procs_nr(); i++) {
			if(i == 0) {
				sprintf(procs_str,"%d",assigned_procs[i]);
			}
			else {
				sprintf(procs_str,"%s %d",procs_str,assigned_procs[i]);
			}
		}
		fprintf(stderr,"[%i] CP: my assigned ranks: %s\n",(*desc)->initial_user_rank, procs_str);	

		i = 0;
		for(; i < assigned_cm_procs_nr(); i++) {
			if(i < assigned_cm_procs_nr() - 1) {
				assert(assigned_procs[i+1] - assigned_procs[i] == (*desc)->cp_size);
			}
		}

		free(assigned_procs);
	}
}

int is_cp_proc(int initial_user_rank) {
	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		if(initial_user_rank % (*desc)->group_size == 0) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	} 
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		if(initial_user_rank < (*desc)->cp_size) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}
}



int is_cm_proc(int initial_user_rank) {
	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		if(initial_user_rank % (*desc)->group_size != 0) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	} 
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		if(initial_user_rank >= (*desc)->cp_size) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}
}

int cp_proc() {
	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		if((*desc)->initial_user_rank % (*desc)->group_size == 0) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	} 
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		if((*desc)->initial_user_rank < (*desc)->cp_size) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}
}

int cm_proc() {
	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		if((*desc)->initial_user_rank % (*desc)->group_size != 0) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	} 
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		if((*desc)->initial_user_rank >= (*desc)->cp_size) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}
}

int is_first_cp_process() {
	if((*desc)->initial_user_rank == 0) {

		return TRUE;
	}
	else {

		return FALSE;
	}
}

int is_first_cm_process() {
	assert(cm_proc());

	if((*desc)->flib_level_rank == 0) {

		return TRUE;
	}
	else {

		return FALSE;
	}
}

int get_first_cm_process_flib_rank() {
	return 0;
}

int first_cp_proc(int rank) {
	if(rank == 0) {
		return TRUE;
	}
	else {
		return FALSE;
	}
}


int first_cm_process(int rank) {
	if(rank == 0) {
		return TRUE;
	}
	else {
		return FALSE;
	}
}

/* TODO: finish these methods. But they might be unnecessary*/

/*int translate_from_world_to_flib_rank(int world_rank) {

	assert(is_cm_proc(world_rank));

	MPI_Group world_group;
	MPI_Group flib_group;

	MPI_Comm_group(MPI_COMM_WORLD, &world_group);
	MPI_Comm_group((*desc)->flib_level_comm, &flib_group);

	int ranks_world = world_rank;
	int ranks_flib = -1;

	MPI_Group_translate_ranks(world_group, 1, &ranks_world, flib_group, &ranks_flib);

	MPI_Group_free(&world_group);
	MPI_Group_free(&flib_group);

	return ranks_flib;
	}*/

/* flib_rank has to be connected with CM process*/
int translate_from_flib_to_world_rank(int flib_rank) {
	int world_rank = -1;

	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		world_rank = flib_rank + 1 + flib_rank / ((*desc)->group_size - 1);
	}
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		world_rank = flib_rank + (*desc)->cp_size;
	}

	assert(flib_rank < (*desc)->cm_size);
	assert(world_rank < (*desc)->initial_user_size);

	return world_rank;
}

//TODO: double check!!!
int translate_from_world_to_flib_rank(int world_rank) {
	int flib_rank = -1;

	if((*desc)->hierarchy_scheme == _FLIB_NONE_HIERARCHY) {
		flib_rank = world_rank - 1 - world_rank / (*desc)->group_size;
	}
	else if((*desc)->hierarchy_scheme == _FLIB_NODES_HIERARCHY) {
		flib_rank = world_rank - (*desc)->cp_size;
	}

	assert(flib_rank < (*desc)->cm_size);
	assert(world_rank < (*desc)->initial_user_size);

	return flib_rank;
}

/*
	 int* translate_from_world_to_flib_rank(int world_rank) {

	 assert(is_cm_proc(world_rank));

	 MPI_Group world_group;
	 MPI_Group flib_group;

	 MPI_Comm_group(MPI_COMM_WORLD, &world_group);
	 MPI_Comm_group((*desc)->flib_level_comm, &flib_group);

	 int ranks_world = world_rank;
	 int ranks_flib = -1;

	 MPI_Group_translate_ranks(world_group, 1, &ranks_world, flib_group, &ranks_flib);

	 MPI_Group_free(&world_group);
	 MPI_Group_free(&flib_group);

	 return ranks_flib;
	 }

	 int translate_from_flib_to_world_rank(int flib_rank) {

	 assert(flib_rank < (*desc)->cm_size);

	 MPI_Group world_group;
	 MPI_Group flib_group;

	 MPI_Comm_group(MPI_COMM_WORLD, &world_group);
	 MPI_Comm_group((*desc)->flib_level_comm, &flib_group);

	 int ranks_world = -1;
	 int ranks_flib = flib_rank;

	 MPI_Group_translate_ranks(flib_group, 1, &ranks_flib, world_group, &ranks_world);

	 MPI_Group_free(&world_group);
	 MPI_Group_free(&flib_group);

	 return ranks_world;
	 }*/
