#include "flib_logging_protocol.h"
#include "flib_debug.h"
#include "flib_markers.h"
#include "flib_cm_proc.h"
#include "flib_counters.h"

//#######################################################################################################################################
// IMPORTANT: in each logging function we use flib level rank 
//#######################################################################################################################################

//#######################################################################################################################################
//various helper functions
//#######################################################################################################################################

int do_logging() {
	if((*desc)->flib_parameters->log_puts == FALSE && (*desc)->flib_parameters->log_gets == FALSE) {
		return FALSE;
	}
	else {
		return TRUE;
	}
}

int self_rank(int target_rank) {
	if((*desc)->flib_level_rank == target_rank) {
		return TRUE;
	}
	else {
		return FALSE;
	}
}

int degenerated() {
	if((*desc)->degenerated == TRUE) {
		return TRUE;
	}
	else {
		return FALSE;
	}
}

//TODO: make a variant in which we get the maximum available amount of available memory for the logging space
//and use this amount while allocating memory for logging space
size_t log_list_size() {
	return (*desc)->flib_parameters->log_list_size;
}

//#######################################################################################################################################
//functions for selecting ranks that have to clear logs, for a given array with log sizes
//#######################################################################################################################################

//TODO: check and verify MAX and ALL
gdsl_queue_t* get_ranks_to_clear(int64_t* array, int array_size) {
	gdsl_queue_t* ranks = malloc(sizeof(gdsl_queue_t));
	*ranks = gdsl_queue_alloc(NULL, NULL, NULL);

	double avg = 0;
	double sum = 0;
	int max_rank = 0;
	int j = 0;

	//TODO: address self-ops!!!!!!!!
	if((*desc)->flib_level_rank == 0) {
		max_rank++;
	}

	for(; j < array_size; j++) {
		if(j == (*desc)->flib_level_rank) {
			continue;
		}
		sum += array[j];
		if(array[j] > array[max_rank]) {
			max_rank = j;
		}
		//debug_message_gen( FALSE,stderr, ")))))))))) local size for PROC [%i]: %i bytes\n",j,array[j]);
	}
	avg = sum / (double)array_size;

	for(j = 0; j < array_size; j++) {
		if(j == (*desc)->flib_level_rank) {
			continue;
		}
		if(( (*desc)->demand_target_selection_heuristics == FLIB_AVERAGE && array[j] >= avg) ||
				( (*desc)->demand_target_selection_heuristics == FLIB_MAX && array[j] == array[max_rank]) ||
				( (*desc)->demand_target_selection_heuristics == FLIB_ALL)) {
			int* new_rank = malloc(sizeof(int));
			*new_rank = j;
			gdsl_queue_insert(*ranks, new_rank);
			//debug_message_gen( FALSE,stderr, ")))))))))) chosen PROC [%i]: %i bytes\n",j,array[j]);
		}
	}
	return ranks;
}

//#######################################################################################################################################
//functions that analyze put logging space and react if the space left if too small
//#######################################################################################################################################

void analyze_put_logging_space(int next_node_offset) {
#ifdef MEASURE_LOGGING_OVERHEADS
	return;
#endif

	gdsl_queue_t* ranks_to_clear = get_ranks_with_huge_put_log_sums();

	if(gdsl_queue_is_empty(*ranks_to_clear) && next_node_offset >= get_remaining_local_space()) {
		//TODO: address it. this exit here causes libdmapp errors
		debug_message_gen_e( FALSE, stderr, " ERROR: no PUT logging space left. yet no ranks to clear found !!!\n");
	}

	while(!gdsl_queue_is_empty(*ranks_to_clear)) {
		int cm_rank_to_clear = *((int*)(gdsl_queue_remove(*ranks_to_clear)));
		int rank_to_clear = translate_from_flib_to_world_rank( cm_rank_to_clear );
		if((*cm_desc)->sent_checkpoint_request[rank_to_clear] == FALSE) {
			MPI_Send(&rank_to_clear, 1, MPI_INT, assigned_cp_proc(rank_to_clear), CM_DEMAND_PUT_CHECKPOINT_REQUEST, (*desc)->initial_user_comm);

			//TODO: previously this whole stuff was AFTER the bracket!!! address
			(*cm_desc)->sent_checkpoint_request[rank_to_clear] = TRUE;
			assert((*cm_desc)->sent_checkpoint_request_epoch[rank_to_clear] == _FLIB_DEMAND_PUT_NOT_SENT);
			(*cm_desc)->sent_checkpoint_request_epoch[rank_to_clear] = (*cm_desc)->local_epoch_counter[cm_rank_to_clear];
			(*cm_desc)->sent_cm_adaptive_requests++;


#ifdef FLIB_DEMAND_DEBUG
			debug_message_gen( FALSE, stderr,
					"*** CM: sent checkpoint request to CP [%i,IN] ([%i,IN] has to take the checkpoint. Reason - PUTS. It will delete %i bytes ). Epoch below which we shall clear the data: %i\n",
					assigned_cp_proc(rank_to_clear), rank_to_clear, ((*cm_desc)->log_space->local_sums)[cm_rank_to_clear], (*cm_desc)->local_epoch_counter[cm_rank_to_clear]);
#endif
		}
	}
	free(ranks_to_clear);
}

gdsl_queue_t* get_ranks_with_huge_put_log_sums() {
	gdsl_queue_t* ranks_to_clear = malloc(sizeof(gdsl_queue_t));
	*ranks_to_clear = gdsl_queue_alloc(NULL, NULL, NULL);

	Log_space* log_space = (*cm_desc)->log_space;
	int64_t remaining_puts_space = get_remaining_local_space();
	double ratio_puts_left = ((double)remaining_puts_space / (double)(*desc)->flib_parameters->log_list_size);

	if((*desc)->degenerated == FALSE && ((*cm_desc)->sent_cm_adaptive_requests == 0) &&
			(ratio_puts_left < (*desc)->flib_parameters->puts_relative_demand_threshold)) {

		assert(remaining_puts_space >= 0 && remaining_puts_space <= (*desc)->flib_parameters->log_list_size);
		ranks_to_clear = get_ranks_to_clear(log_space->local_sums, (*desc)->cm_size);
	}
	return ranks_to_clear;
}

int get_remaining_local_space() {
	return (*desc)->flib_parameters->log_list_size - (*cm_desc)->log_space->next_free_local_offset;
}

//#######################################################################################################################################
//functions that analyze get logging space and react if the space left is too small
//#######################################################################################################################################

void analyze_get_logging_space(int new_log_size) {
#ifdef MEASURE_LOGGING_OVERHEADS
	return;
#endif

	if(!do_logging()) {return;}
	if(degenerated()) {return;}

	gdsl_queue_t* ranks_to_clear = get_ranks_with_huge_get_log_sums();

	while(!gdsl_queue_is_empty(*ranks_to_clear)) {
		int cm_rank_to_clear = *((int*)(gdsl_queue_remove(*ranks_to_clear)));
		int rank_to_clear = translate_from_flib_to_world_rank( cm_rank_to_clear );

		//TODO: make the size smaller (cm_size)
		if((*cm_desc)->sent_checkpoint_request[rank_to_clear] == FALSE) {
			MPI_Send(&rank_to_clear, 1, MPI_INT, assigned_cp_proc(rank_to_clear), CM_DEMAND_GET_CHECKPOINT_REQUEST, (*desc)->initial_user_comm);

			//TODO: check if we want to change this to separate tables for puts/gets
			(*cm_desc)->sent_checkpoint_request[rank_to_clear] = TRUE;

			//we keep separate counters for, e.g., statistical purposes
			(*cm_desc)->sent_cm_get_demand_requests++;

#ifdef FLIB_DEMAND_DEBUG
			debug_message_gen( FALSE, stderr,
					"&&& CM: sent GET checkpoint request to CP [%i,IN] ([%i,IN] has to take the checkpoint. Reason - GETS).\n",
					assigned_cp_proc(rank_to_clear), rank_to_clear);
#endif
		}
	}
	free(ranks_to_clear);
}

gdsl_queue_t* get_ranks_with_huge_get_log_sums() {
	gdsl_queue_t* ranks_to_clear = malloc(sizeof(gdsl_queue_t));
	*ranks_to_clear = gdsl_queue_alloc(NULL, NULL, NULL);

	Log_space* log_space = (*cm_desc)->log_space;
	int64_t remaining_gets_space = get_remaining_global_space((*desc)->flib_level_rank);
	double ratio_gets_left = ((double)remaining_gets_space / (double)(*desc)->flib_parameters->log_list_size);

	if(((*cm_desc)->sent_cm_get_demand_requests == 0) && (ratio_gets_left < (*desc)->flib_parameters->gets_relative_demand_threshold)) {

		assert(remaining_gets_space >= 0 && remaining_gets_space <= (*desc)->flib_parameters->log_list_size);

		int64_t* faked_var = malloc(sizeof(int64_t) * (*desc)->cm_size );
		int64_t* gets_sums = malloc(sizeof(int64_t) * (*desc)->cm_size );

		//foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_logs_control_win));
		foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_list));
		//ranks_to_clear = get_ranks_to_clear(log_space->global_logs_control_data, (*desc)->cm_size);

		//TODO: make this more memory efficient (divide this control window)
		foMPI_Get_accumulate(&faked_var[0], (*desc)->cm_size, MPI_INT64_T, &gets_sums[0], (*desc)->cm_size, MPI_INT64_T, (*desc)->flib_level_rank, 0, (*desc)->cm_size, MPI_INT64_T,
				foMPI_NO_OP, *(log_space->global_logs_control_win));
		foMPI_Win_flush((*desc)->flib_level_rank, *((*cm_desc)->log_space->global_logs_control_win));

		ranks_to_clear = get_ranks_to_clear(gets_sums, (*desc)->cm_size);

		foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_list));
	}
	return ranks_to_clear;
}

int get_remaining_global_space(int rank) {
	int64_t size = -1;
	int64_t faked_var = 0;
	Log_space* log_space = (*cm_desc)->log_space;

	//foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_logs_control_win));
	//foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_list));

	foMPI_Get_accumulate(&faked_var, 1, MPI_INT64_T, &size, 1, MPI_INT64_T, rank, (*desc)->cm_size, 1, MPI_INT64_T, foMPI_NO_OP,
			*(log_space->global_logs_control_win));
	foMPI_Win_flush(rank, *(log_space->global_logs_control_win));

	//foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_logs_control_win));
	//foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_list));

	return (*desc)->flib_parameters->log_list_size - size;
}

//#######################################################################################################################################
// functions for logging a put
//#######################################################################################################################################

void finalize_logging_put(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype,int target_rank, MPI_Aint target_disp, 
		int target_count, MPI_Datatype target_datatype, MPI_Op operation, Operation_type op_type) { 
	Log_space* log_space = (*cm_desc)->log_space;

	int next_free = log_space->next_free_local_offset;	

	// get the size of the datatype
	int origin_type_size = -1;
	MPI_Type_size(origin_datatype, &origin_type_size);
	assert(origin_type_size > 0);

	// I calculate the payload as origin_count * origin_datatype. I assume that target_datatype * target_count is the same
	int payload_size = origin_count * origin_type_size;   
	int next_node_offset = sizeof(Common_params) + payload_size;

	if(op_type == PUT) {
		next_node_offset += sizeof(Put_get_args);
	}
	else if(op_type == ACCUMULATE) {
		next_node_offset += sizeof(Acc_args);
	}

	// get the current epoch counter (local, as we're recording a put).
	// I can get this value here (outside a critical section) because
	// this is an active counter, incremented only by me, no other proc 
	int epoch_nr = (*cm_desc)->local_epoch_counter[target_rank];

	double ratio_left = ((double)get_remaining_local_space() / (double)(*desc)->flib_parameters->log_list_size);
#ifdef FLIB_LOGGING_PUTS_VERBOSE_DEBUG
	debug_message_gen( FALSE, stderr,
			"+++ CM: Begin to log a PUT. Remaing space: %i bytes (%.2f%%). New PUT log will occupy: %i bytes. Current superstep (fence): %i. EA[target]: %i\n",
			get_remaining_local_space(), ratio_left*100, next_node_offset, (*cm_desc)->superstep, epoch_nr);
#endif

	if(next_node_offset > (*desc)->flib_parameters->log_list_size) {
		debug_message_gen_e( TRUE, stderr, "########### the PUT to be logged is bigger that the whole put logging space!!! increase the log size; exiting...\n");
	}
	if(ratio_left < (*desc)->flib_parameters->puts_relative_demand_threshold && (*cm_desc)->superstep == 0 && (*desc)->app_type == FLIB_APP_BLK) {
		debug_message_gen_e( TRUE, stderr, "########### more puts to be logged in one superstep than the whole put logging space!!! increase the log size; exiting...\n");
	}

#ifdef MEASURE_LOGGING_OVERHEADS
	assert(ratio_left > 0.1);
#else
	//First, we check if the space left is lower than the 'dangerous' threshold	
	//if yes, we send the demand checkpoint
	analyze_put_logging_space(next_node_offset); 
	//analyze_get_logging_space(0);	//TODO: double check if this is unnecessary

	//Then, if the space left is actually not enough to perform the logging, we
	//wait until we receive a notification that we can delete part of the logs

	while( next_node_offset >= get_remaining_local_space() ) {
		//TODO: make sure we react correctly to the coordinated checkpoint
		if((*cm_desc)->sent_cm_adaptive_requests == 0 && next_node_offset >= get_remaining_local_space()) {
			analyze_put_logging_space(next_node_offset);
		}
		cm_process_notifications();
	}
#endif

	// get the offset of the beginning of the free space in the log list
	char* target_ptr = &((char*)(log_space->local_list_data))[next_free];

	Common_params common_params;
	common_params.epoch_nr = epoch_nr;
	common_params.target_rank = target_rank;
	common_params.next_node_offset = next_node_offset;
	common_params.op_type = op_type;
	common_params.GNC = (*cm_desc)->GNC_cnt;
	common_params.GC = (*cm_desc)->GC_cnt;

	//TODO: do the SC counter
	common_params.SC = (*cm_desc)->SC_cnt;

#ifndef MEASURE_LOGGING_OVERHEADS
	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->local_list));
#endif

	// log the common params
	memcpy(target_ptr, &common_params, sizeof(Common_params));
	target_ptr += sizeof(Common_params);

	if(op_type == PUT) {
		Put_get_args put_get_args;
		put_get_args.origin_count = origin_count;
		put_get_args.origin_datatype = origin_datatype;
		put_get_args.target_disp = target_disp;
		put_get_args.target_count = target_count;
		put_get_args.target_datatype = target_datatype;
	
		common_params.combining = FALSE;	

		// log the put-specific params
		memcpy(target_ptr, &put_get_args, sizeof(Put_get_args));
		target_ptr += sizeof(Put_get_args);
	} else if(op_type == ACCUMULATE) {
		Acc_args acc_args;
		acc_args.origin_count = origin_count;
		acc_args.origin_datatype = origin_datatype;
		acc_args.target_disp = target_disp;
		acc_args.target_count = target_count;
		acc_args.target_datatype = target_datatype;
		acc_args.operation = operation;

		common_params.combining = TRUE;

		// log the accumulate-specific params
		memcpy(target_ptr, &acc_args, sizeof(Acc_args));
		target_ptr += sizeof(Acc_args);
	}

	// log the payload
	memcpy(target_ptr, origin_addr, payload_size);
	target_ptr += payload_size;

	// update the logging info
	log_space->next_free_local_offset += next_node_offset;
	log_space->number_of_nodes_local += 1;
	(log_space->local_sums)[target_rank] += next_node_offset;

	foMPI_Win_sync(*(*cm_desc)->log_space->local_list_control);
	foMPI_Win_sync(*(*cm_desc)->log_space->local_list_size);

#ifdef FLIB_LOGGING_PUTS_DEBUG
	//testing log
	foMPI_Win_flush((*desc)->flib_level_rank, *(log_space->local_list));

	__log_put( origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, operation, op_type, FALSE);
	__test_consistency_of_puts(FALSE);
#endif

#ifdef FLIB_LOGGING_PUTS_VERBOSE_DEBUG
	ratio_left = ((double)get_remaining_local_space() / (double)(*desc)->flib_parameters->log_list_size);
	debug_message_gen( FALSE, stderr, "+++ CM: PUT logged. Remaing space: %i bytes (%.2f%%)\n", get_remaining_local_space(), ratio_left*100);
#endif

#ifndef MEASURE_LOGGING_OVERHEADS
	foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->local_list));
#endif
}

//#######################################################################################################################################
// functions for logging a get
//#######################################################################################################################################

void finalize_logging_get(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count,
		MPI_Datatype target_datatype) {
	Log_space* log_space = (*cm_desc)->log_space;

	Local_get_info* get_info = malloc( sizeof(Local_get_info) );
	assert(get_info != NULL);	

	get_info->origin_addr = origin_addr; 
	get_info->origin_count = origin_count;
	get_info->origin_datatype = origin_datatype;
	get_info->target_rank = target_rank; //TODO: probably not necessary
	get_info->target_disp = target_disp;
	get_info->target_count = target_count;
	get_info->target_datatype = target_datatype;

	gdsl_queue_insert(log_space->gets_to_be_logged[target_rank], get_info);

#ifdef FLIB_LOGGING_GETS_VERBOSE_DEBUG
	debug_message_gen( FALSE, stderr, "&&& CM: GET initially logged (target rank: [%i,IN])\n", translate_from_flib_to_world_rank(target_rank));
#endif
}

void log_outstanding_operations(int rank) {
	//if logging degenerated, we had already cleared the data to be logged, and we can exit now
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	Log_space* log_space = (*cm_desc)->log_space;

	while(!gdsl_queue_is_empty(log_space->gets_to_be_logged[rank])) {
		Local_get_info* get_info = (Local_get_info*) gdsl_queue_remove(log_space->gets_to_be_logged[rank]);
		log_outstanding_get(get_info);
	}
}

void log_all_outstanding_operations() {
	//if logging degenerated, we have already cleared the data to be logged
	//and we skip this function here
	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	Log_space* log_space = (*cm_desc)->log_space;

	int i = 0;
	//TODO: check this size; previously: initial.
	for(; i < (*desc)->cm_size; i++) {
		//TODO: delete after checking...
		/*if(is_cp_proc((*desc)->initial_user_size)) {
			assert(gdsl_queue_is_empty(log_space->gets_to_be_logged[i]) == TRUE);
			}*/

		while(!gdsl_queue_is_empty(log_space->gets_to_be_logged[i])) {
			Local_get_info* get_info = (Local_get_info*) gdsl_queue_remove(log_space->gets_to_be_logged[i]);
			log_outstanding_get(get_info);
		}
	}
}

void log_outstanding_get(Local_get_info* get_info) {
	Common_params params;
	Put_get_args args;

	Log_space* log_space = (*cm_desc)->log_space;

	// TODO: make sure that this epoch nr is 100% correct
	int epoch_nr = (*cm_desc)->local_epoch_counter[get_info->target_rank];
	int origin_type_size = -1;
	MPI_Type_size(get_info->origin_datatype, &origin_type_size);
	assert(origin_type_size > 0);

	int payload_size = get_info->origin_count * origin_type_size;

	int next_node_offset = + sizeof(Common_params)    // operation type 
		+ sizeof(Put_get_args)     // arguments size 
		+ payload_size;             // payload size

	int64_t next_node_size = sizeof(Common_params) + sizeof(Put_get_args) + payload_size;
	int64_t nodes_nr_addition = 1;

	params.epoch_nr = epoch_nr;

	//TODO: do it more elegantly
	params.target_rank = (*desc)->flib_level_rank;//get_info->target_rank;
	params.next_node_offset = next_node_offset;
	params.op_type = GET;

	params.GNC = (*cm_desc)->GNC_cnt;
	params.GC = (*cm_desc)->GC_cnt;

	//TODO: do the SC counter
	params.SC = (*cm_desc)->SC_cnt;
	params.combining = FALSE;

	args.origin_count = get_info->origin_count;
	args.origin_datatype = get_info->origin_datatype;
	args.target_disp = get_info->target_disp;
	args.target_count = get_info->target_count;
	args.target_datatype = get_info->target_datatype;

	// TODO: make it faster by using a smarter strategy (than locks)
	if((*desc)->logging_gets_scheme == FLIB_LOGGING_GETS_LOCK) {
		/* previously I used an atomic here with no lock to get the initial value of remote empty space
			 it could lead to some data corruption or more complex solutions. 
			 Thus we have a lock around the whole section */

		int64_t size = -1;
		int64_t faked_var = -1;
		double ratio_left = 0;
		int remaining_space = 0;

		int h = 0;	
		do {
			h++;
			if(h > 10) {
				debug_message_gen_e(FALSE, stderr, "Pomoz mi, Panie Moj...\n");
			}
			foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, get_info->target_rank, 0, *(log_space->global_list)); 

			foMPI_Get_accumulate(&faked_var, 1, MPI_INT64_T, &size, 1, MPI_INT64_T, get_info->target_rank, (*desc)->cm_size, 1, MPI_INT64_T, foMPI_NO_OP, *((*cm_desc)->log_space->global_logs_control_win));
			foMPI_Win_flush(get_info->target_rank, *((*cm_desc)->log_space->global_logs_control_win));//TODO: check it, previously it was just 'rank'...

			remaining_space = (*desc)->flib_parameters->log_list_size - size;
			ratio_left = ((double)remaining_space)/(double)(*desc)->flib_parameters->log_list_size;

#ifdef FLIB_LOGGING_GETS_VERBOSE_DEBUG
			debug_message_gen( FALSE, stderr,
					"+++ CM: Begin to log a GET. Remaing space: %i bytes (%.2f%%). New GET log will occupy: %i bytes. Current superstep (fence): %i. EA[target]: %i\n",
					remaining_space, ratio_left*100, next_node_offset, (*cm_desc)->superstep, epoch_nr);
#endif

			if(next_node_offset > (*desc)->flib_parameters->log_list_size) {
				foMPI_Win_unlock(get_info->target_rank, *(log_space->global_list));
				debug_message_gen_e( TRUE, stderr, "########### the GET to be logged is bigger that the whole get logging space!!! increase the log size; exiting...\n");
			}
			if(ratio_left < (*desc)->flib_parameters->gets_relative_demand_threshold && (*cm_desc)->superstep == 0 && (*desc)->app_type == FLIB_APP_BLK) {
				foMPI_Win_unlock(get_info->target_rank, *(log_space->global_list));
				debug_message_gen( TRUE, stderr, "########### more gets to be logged in one superstep (current nr: %i) than the whole get logging space!!! increase the log size; exiting...\n", (*cm_desc)->superstep);
			}

			if(next_node_offset >= remaining_space) {
				foMPI_Win_unlock(get_info->target_rank, *(log_space->global_list));
			}

			cm_process_notifications();
		} while(next_node_offset >= remaining_space);

		int64_t remote_next_free = -1;
		foMPI_Get_accumulate(&next_node_size, 1, MPI_INT64_T, &remote_next_free, 1, MPI_INT64_T, get_info->target_rank, (*desc)->cm_size, 1, MPI_INT64_T, foMPI_SUM,
				*(log_space->global_logs_control_win));
		foMPI_Win_flush(get_info->target_rank, *(log_space->global_logs_control_win));

		int target_disp = remote_next_free;

		foMPI_Put(&params, sizeof(Common_params), MPI_CHAR, get_info->target_rank, target_disp, sizeof(Common_params), MPI_CHAR, *(log_space->global_list));
		target_disp += sizeof(Common_params);

		foMPI_Put(&args, sizeof(Put_get_args), MPI_CHAR, get_info->target_rank, target_disp, sizeof(Put_get_args), MPI_CHAR, *(log_space->global_list));
		target_disp += sizeof(Put_get_args);

		foMPI_Put((char*)(get_info->origin_addr), payload_size, MPI_CHAR, get_info->target_rank, target_disp, payload_size, MPI_CHAR, *(log_space->global_list));
		target_disp += payload_size;

		foMPI_Accumulate(&next_node_size, 1, MPI_INT64_T, get_info->target_rank, (*desc)->flib_level_rank, 1, MPI_INT64_T, foMPI_SUM, *(log_space->global_logs_control_win));
		foMPI_Accumulate(&nodes_nr_addition, 1, MPI_INT64_T, get_info->target_rank, (*desc)->cm_size + 1, 1, MPI_INT64_T, foMPI_SUM, *(log_space->global_logs_control_win));
		foMPI_Win_flush(get_info->target_rank, *(log_space->global_logs_control_win));

#ifdef FLIB_LOGGING_GETS_DEBUG
		__log_get(get_info->origin_addr, get_info->origin_count, get_info->origin_datatype,
				get_info->target_rank, get_info->target_disp, get_info->target_count, get_info->target_datatype, epoch_nr, FALSE);
#endif

#ifdef FLIB_LOGGING_GETS_VERBOSE_DEBUG
		int64_t left = (*desc)->flib_parameters->log_list_size - remote_next_free;
		ratio_left = ((double)left / (double)(*desc)->flib_parameters->log_list_size);
		debug_message_gen( FALSE, stderr, "&&& CM: GET logged. Remaing space at CM [%i]: %lli bytes (%.2f%%)\n",
				get_info->target_rank, left, ratio_left*100);//TODO: previously: world_target_rank!
#endif

		// TODO: make sure this place for this operation is safe
		(log_space->procs_communicated_using_gets)[get_info->target_rank] = TRUE;

		foMPI_Win_unlock(get_info->target_rank, *(log_space->global_list));

		clear_M_marker(get_info->target_rank);
	}
	else {
		debug_message_gen_e( TRUE, stderr, "&&& CM: ERROR: Unknown gets logging policy\n");
	}
}

//#######################################################################################################################################
//functions that initiate the logging of each of the RMA communication functions
//#######################################################################################################################################

void log_put(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype) {
#ifndef MEASURE_LOGGING_OVERHEADS
	if(degenerated()) {return;}
	if(!do_logging()) {return;}
#endif
	if(self_rank(target_rank)) {return;}

	if((*desc)->flib_parameters->log_puts == TRUE) {
		finalize_logging_put(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, foMPI_REPLACE, PUT);
	}
}

void log_accumulate(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count,
		MPI_Datatype target_datatype, MPI_Op operation) {
#ifndef MEASURE_LOGGING_OVERHEADS
	if(degenerated()) {return;}
	if(!do_logging()) {return;}
#endif
	if(self_rank(target_rank)) {return;}

	if((*desc)->flib_parameters->log_puts == TRUE) {
		finalize_logging_put(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, operation, ACCUMULATE);
	}
}

void log_get(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype) {
#ifndef MEASURE_LOGGING_OVERHEADS
	if(degenerated()) {return;}
	if(!do_logging()) {return;}
#endif
	if(self_rank(target_rank)) {return;}

	if((*desc)->flib_parameters->log_gets == TRUE) {
		finalize_logging_get(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype);
	}
}

void log_get_accumulate(const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, void *result_addr, int result_count, MPI_Datatype result_datatype,
		int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype, MPI_Op op) {
#ifndef MEASURE_LOGGING_OVERHEADS
	if(degenerated()) {return;}
	if(!do_logging()) {return;}
#endif
	if(self_rank(target_rank)) {return;}

	if((*desc)->flib_parameters->log_puts == TRUE) {
		finalize_logging_put(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, op, ACCUMULATE);
	}

	if((*desc)->flib_parameters->log_gets == TRUE) {
		finalize_logging_get(result_addr, result_count, result_datatype, target_rank, target_disp, target_count, target_datatype);
	}
}

void log_fetch_and_op(const void *origin_addr, void *result_addr, MPI_Datatype datatype, int target_rank, MPI_Aint target_disp, MPI_Op op) {
#ifndef MEASURE_LOGGING_OVERHEADS
	if(degenerated()) {return;}
	if(!do_logging()) {return;}
#endif
	if(self_rank(target_rank)) {return;}

	if((*desc)->flib_parameters->log_puts == TRUE) {
		finalize_logging_put(origin_addr, 1, datatype, target_rank, target_disp, 1, datatype, op, ACCUMULATE);
	}

	if((*desc)->flib_parameters->log_gets == TRUE) {
		finalize_logging_get(result_addr, 1, datatype, target_rank, target_disp, 1, datatype);
	}
}

void log_compare_and_swap(const void *origin_addr, const void *compare_addr, void *result_addr, MPI_Datatype datatype, int target_rank, MPI_Aint target_disp) {
#ifndef MEASURE_LOGGING_OVERHEADS
	if(degenerated()) {return;}
	if(!do_logging()) {return;}
#endif
	if(self_rank(target_rank)) {return;}

	if((*desc)->flib_parameters->log_puts == TRUE) {
		finalize_logging_put(origin_addr, 1, datatype, target_rank, target_disp, 1, datatype, foMPI_REPLACE, PUT);
	}

	if((*desc)->flib_parameters->log_gets == TRUE) {
		finalize_logging_get(result_addr, 1, datatype, target_rank, target_disp, 1, datatype);
	}
}

//#######################################################################################################################################
//Functions for clearing local & global spaces
//#######################################################################################################################################

void clear_local_logging_space(int rank_to_clear, int epoch_nr) {
#ifdef MEASURE_LOGGING_OVERHEADS
	debug_message_gen_e( TRUE, stderr, "MMMMM: no space left in local logging space!!! \n");
	return;
#endif

	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	Log_space* log_space = (*cm_desc)->log_space;

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->local_list));

	int space_before_clear = get_remaining_local_space();

#ifdef FLIB_LOCAL_CLEAR_VERBOSE_DEBUG
	debug_message_gen_e( FALSE, stderr, "+++ CM: begin clearing local LOGS\n");
#endif

#ifdef FLIB_LOCAL_CLEAR_DEBUG
	__test_consistency_of_puts(TRUE);
#endif

	// get the beginning addres
	char* target_ptr = ((char*)(log_space->local_list_data));
	char* local_logs_begin = ((char*)(log_space->local_list_data));
	int nodes_analyzed = 0;
	int nodes_deleted = 0;
	int bytes_from_beginning = 0;

	int nodes_to_copy = 0;
	int nodes_to_copy_offsets[log_space->number_of_nodes_local];
	int nodes_to_copy_sizes[log_space->number_of_nodes_local];

	int rank_to_clear_cm = translate_from_world_to_flib_rank(rank_to_clear);

	for(nodes_analyzed = 0; nodes_analyzed < log_space->number_of_nodes_local; nodes_analyzed++) {
		Common_params params;
		memcpy(&params, target_ptr, sizeof(Common_params));

		//we assume here that the user correctly uses a new value of rank: fLIB user rank
		if(params.epoch_nr >= epoch_nr || params.target_rank != rank_to_clear_cm) {
			nodes_to_copy_offsets[nodes_to_copy] = bytes_from_beginning;
			nodes_to_copy_sizes[nodes_to_copy] = params.next_node_offset;
			nodes_to_copy++;
		}

		bytes_from_beginning += params.next_node_offset;
		target_ptr += params.next_node_offset;
	}

	target_ptr = ((char*)(log_space->local_list_data));	
	int new_next_free_local_offset = 0;

	int i = 0;
	for(i = 0; i < nodes_to_copy; i++) {
		memmove(target_ptr, local_logs_begin + nodes_to_copy_offsets[i], nodes_to_copy_sizes[i]);
		target_ptr += nodes_to_copy_sizes[i];
		new_next_free_local_offset += nodes_to_copy_sizes[i];
	}

	log_space->next_free_local_offset = new_next_free_local_offset;
	log_space->number_of_nodes_local = nodes_to_copy;

	int remaining_space = get_remaining_local_space();
	(log_space->local_sums)[rank_to_clear_cm] -= (remaining_space - space_before_clear);
	assert(remaining_space - space_before_clear >= 0);	//TODO: address the scenario where we get 0 more bytes of free space...
	assert(remaining_space >= 0);
	assert((log_space->local_sums)[rank_to_clear_cm] >= 0);

#ifdef FLIB_LOCAL_CLEAR_DEBUG
	foMPI_Win_flush((*desc)->flib_level_rank, *(log_space->local_list));

	__delete_put_testing_logs(rank_to_clear_cm, epoch_nr, FALSE);
	__test_consistency_of_puts(TRUE);
#endif

	foMPI_Win_sync(*(*cm_desc)->log_space->local_list_control);
	foMPI_Win_sync(*(*cm_desc)->log_space->local_list_size);
	
	foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->local_list));

#ifdef FLIB_LOCAL_CLEAR_VERBOSE_DEBUG
	double ratio_left = ((double)remaining_space / (double)(*desc)->flib_parameters->log_list_size);
	debug_message_gen( FALSE, stderr, "&&& Local logs deleted. Remaining space: %i (%.2f%%). CLEARED SPACE: %i bytes. Nodes to copy: %i. Nodes analyzed: %i\n",
			get_remaining_local_space(), ratio_left*100, remaining_space - space_before_clear, nodes_to_copy, nodes_analyzed);
#endif
}

void clear_global_logging_space(int rank_to_clear, int epoch_nr) {
#ifdef MEASURE_LOGGING_OVERHEADS
	debug_message_gen_e( TRUE, stderr, "MMMMM: no space left in global logging space!!! \n");
	return;
#endif

	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	Log_space* log_space = (*cm_desc)->log_space;

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_list));
	//foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_logs_control_win));

#ifdef FLIB_GLOBAL_CLEAR_VERBOSE_DEBUG
	debug_message_gen( FALSE, stderr, "&&& CM: begin clearing local >>GET<< LOGS, rank to clear: [%i,CM], epoch to clear: %i\n", rank_to_clear, epoch_nr);
#endif

	// get the beginning addres
	char* target_ptr = ((char*)(log_space->global_list_data));
	char* global_logs_begin = ((char*)(log_space->global_list_data));
	int nodes_analyzed = 0;
	int nodes_deleted = 0;
	int bytes_from_beginning = 0;

	int64_t nodes_to_copy = 0;

	int64_t faked_var;	
	//int64_t number_of_nodes_global;
	//currently we need to use atomics even when updating these values locally
	//we don't have them protected by locks anymore
	int64_t number_of_nodes_global = (log_space->global_logs_control_data)[(*desc)->cm_size + 1];

	//foMPI_Get_accumulate(&faked_var, 1, MPI_INT64_T, &number_of_nodes_global, 1, MPI_INT64_T, (*desc)->flib_level_rank, (*desc)->cm_size, 1, MPI_INT64_T, foMPI_NO_OP,
	//      *(log_space->global_logs_control_win));
	//foMPI_Win_flush((*desc)->flib_level_rank, *(log_space->global_logs_control_win));

	int nodes_to_copy_offsets[number_of_nodes_global];
	int nodes_to_copy_sizes[number_of_nodes_global];

	for(nodes_analyzed = 0; nodes_analyzed < number_of_nodes_global; nodes_analyzed++) {
		Common_params params;
		memcpy(&params, target_ptr, sizeof(Common_params));

		if(params.epoch_nr >= epoch_nr || params.target_rank != rank_to_clear) {
			nodes_to_copy_offsets[nodes_to_copy] = bytes_from_beginning;
			nodes_to_copy_sizes[nodes_to_copy] = params.next_node_offset;
			nodes_to_copy++;
		}

		bytes_from_beginning += params.next_node_offset;
		target_ptr += params.next_node_offset;
	}

	target_ptr = ((char*)(log_space->global_list_data));	
	int64_t new_next_free_global_offset = 0;

	int i = 0;
	for(i = 0; i < nodes_to_copy; i++) {
		memmove(target_ptr, global_logs_begin + nodes_to_copy_offsets[i], nodes_to_copy_sizes[i]);
		target_ptr += nodes_to_copy_sizes[i];
		new_next_free_global_offset += nodes_to_copy_sizes[i];
	}

	//foMPI_Accumulate(&new_next_free_global_offset, 1, MPI_INT64_T, (*desc)->flib_level_rank, (*desc)->cm_size, 1, MPI_INT64_T, foMPI_REPLACE, *(log_space->global_logs_control_win));
	//foMPI_Accumulate(&nodes_to_copy, 1, MPI_INT64_T, (*desc)->flib_level_rank, (*desc)->cm_size+1, 1, MPI_INT64_T, foMPI_REPLACE, *(log_space->global_logs_control_win));

	//currently we need to use atomics even when updating these values locally
	//we don't have them protected by locks anymore
	(log_space->global_logs_control_data)[(*desc)->cm_size] = new_next_free_global_offset;
	(log_space->global_logs_control_data)[(*desc)->cm_size + 1] = nodes_to_copy;

#ifdef FLIB_GLOBAL_CLEAR_DEBUG
	//foMPI_Win_flush((*desc)->flib_level_rank, *(log_space->global_list));

	//__delete_get_testing_logs(rank_to_clear, epoch_nr, TRUE);
#endif

#ifdef FLIB_GLOBAL_CLEAR_VERBOSE_DEBUG
	int remaining_space = get_remaining_global_space((*desc)->flib_level_rank);
	double ratio_left = ((double)remaining_space / (double)(*desc)->flib_parameters->log_list_size);
	debug_message_gen(FALSE, stderr, "&&& >>GET<< logs deleted (rank to clear: %i, epoch to clear: %i). Remaining space: %i (%.2f%%)\n",
			rank_to_clear, epoch_nr, get_remaining_global_space((*desc)->flib_level_rank), ratio_left*100);
#endif

	foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_list));
	//foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_logs_control_win));
}

void clear_all_logs() {
#ifdef MEASURE_LOGGING_OVERHEADS
	debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "MMMMM: do not clear all logs (measure logging overheads) \n");
	return;
#endif

	if(degenerated()) {return;}
	if(!do_logging()) {return;}

	Log_space* log_space = (*cm_desc)->log_space;

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->local_list));
	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_list));
	//foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_logs_control_win));

	memset(log_space->local_list_data, 0, (*desc)->flib_parameters->log_list_size * sizeof(char) );
	memset(log_space->global_list_data, 0, (*desc)->flib_parameters->log_list_size * sizeof(char) );

	log_space->next_free_local_offset = 0;
	log_space->number_of_nodes_local = 0;
	memset(log_space->local_sums, 0, sizeof(int) * (*desc)->cm_size);
	memset(log_space->global_logs_control_data, 0, sizeof(int64_t) * ( (*desc)->cm_size + 2));

	// TODO: make sure it's safe place for this operation 
	int i = 0;
	for(; i < (*desc)->cm_size; i++) {
		(log_space->procs_communicated_using_gets)[i] = FALSE;
	}

	gdsl_queue_flush(*(log_space->gets_to_be_logged));

	//foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_logs_control_win));
	foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_list));
	foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->local_list));

#if defined(FLIB_LOCAL_CLEAR_DEBUG) || defined(FLIB_GLOBAL_CLEAR_DEBUG)
	__delete_all_testing_logs();
	__test_consistency_of_puts(TRUE);
#endif

#if defined(FLIB_LOCAL_CLEAR_VERBOSE_DEBUG) || defined(FLIB_GLOBAL_CLEAR_VERBOSE_DEBUG)
	debug_message_gen_e( FALSE, stderr, "&&& CM: All LOGS cleared\n");
#endif
}


//TODO: integrate 'smartly' the necessary checkpoints into the protocol
//code for NECESSARY checkpoints

/*if(ratio_left <= (*desc)->flib_parameters->gets_relative_demand_threshold) {
	int world_target_rank = translate_from_flib_to_world_rank(get_info->target_rank);

	if((*cm_desc)->sent_necessary_checkpoint_done[get_info->target_rank] == FALSE) {

#ifdef FLIB_DEMAND_DEBUG
debug_message_gen( FALSE, stderr, "&&& CM: Rank [%i,IN] has little space for GETS. Remaining space: %lli (%.2f%%). I take a CKPT and send the clear req\n",
world_target_rank, remaining_space, ratio_left*100);
#endif	

int* data = take_demand_checkpoint(get_info->target_rank);
int data_to_send[2];
data_to_send[0] = data[0];
data_to_send[1] = get_info->target_rank;

MPI_Send(&data_to_send[0], 2, MPI_INT, my_assigned_cp_proc(), CM_NECESSARY_CHECKPOINT_DONE, (*desc)->initial_user_comm);
(*cm_desc)->sent_necessary_checkpoint_done[get_info->target_rank] = TRUE;
(*cm_desc)->sent_cm_necessary_requests++;
free(data);

#ifdef FLIB_DEMAND_DEBUG
debug_message_gen( FALSE, stderr, "&&& CM: Necessary checkpoint done, sent confirmation to CP [%i,IN] (to CM [%i,IN])\n",
my_assigned_cp_proc(), world_target_rank);
#endif
}

if(next_node_offset >= remaining_space) {
foMPI_Win_unlock(get_info->target_rank, *(log_space->global_list));

while((*cm_desc)->sent_necessary_checkpoint_done[get_info->target_rank] == TRUE) {
cm_process_notifications();
}
continue;
}
else {
break;
}
}*/


