#ifndef _FLIB_LOGGING_PROTOCOL_H_
#define _FLIB_LOGGING_PROTOCOL_H_

#include "flib.h"
#include "flib_cm_proc.h"

gdsl_queue_t* get_ranks_to_clear(int64_t* array, int array_size);
int get_max_index(int* array, int array_size);
size_t log_list_size();
int arguments_size(Operation_type op_type);

void inc_global_epoch_counter(int rank);
void inc_local_epoch_counter(int rank);

void inc_global_epoch_counters();
void inc_local_epoch_counters();

void log_put(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype);
void log_accumulate(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count,
										MPI_Datatype target_datatype, MPI_Op operation);
void finalize_logging_put(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype,int target_rank, MPI_Aint target_disp,
                          int target_count, MPI_Datatype target_datatype, MPI_Op operation, Operation_type op_type);

void finalize_logging_get(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype,int target_rank, MPI_Aint target_disp,
                          int target_count, MPI_Datatype target_datatype);

void log_get(const void* origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype);

void log_get_accumulate(const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, void *result_addr, int result_count, MPI_Datatype result_datatype,
                         int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype, MPI_Op op);
void log_fetch_and_op(const void *origin_addr, void *result_addr, MPI_Datatype datatype, int target_rank, MPI_Aint target_disp, MPI_Op op);
void log_compare_and_swap(const void *origin_addr, const void *compare_addr, void *result_addr, MPI_Datatype datatype, int target_rank, MPI_Aint target_disp);

int do_logging();
int degenerated();
int self_rank(int target_rank);

void analyze_remaining_local_space(int next_node_offset);
int get_remaining_local_space();
int get_remaining_global_space();
int get_remaining_global_space_remote(int rank);

gdsl_queue_t* get_ranks_with_huge_put_log_sums();
gdsl_queue_t* get_ranks_with_huge_get_log_sums();

void analyze_logging_space();
void analyze_put_logging_space(int);
void analyze_get_logging_space(int);

void log_operation(Operation_type op_type);

void log_outstanding_operations(int target_rank);
void log_all_outstanding_operations();

void log_outstanding_get(Local_get_info* get_info);

void free_outstanding_operations(int target_rank);
void free_all_outstanding_operations();

void clear_local_logging_space(int rank_to_clear, int epoch_to_clear);
void clear_global_logging_space(int rank_to_clear, int epoch_to_clear);

void clear_all_logs();
#endif
