#include "flib_recovery.h"	
#include "flib_logging_protocol.h"

//#######################################################################################################################################
//A routine called by a crashed process to begin the recovery
//#######################################################################################################################################

//TODO: address the issue of receving other msgs (e.g. demand checkpoint requests) while being here.
//now I assume we can receive only request and checkpoint data
//TODO: solve the problems connected with communicator state after a crash. In the current version we don't care about it

void cm_global_recovery() {
	int assigned_cp_proc = my_assigned_cp_proc();

	//First, we notify out CP tghat we need to recover
	//The CP will get the checkpoint data from other procs and send it to us

	MPI_Send(NULL, 0, MPI_CHAR, assigned_cp_proc, CM_PROC_GLOBAL_RECOVERY, (*desc)->initial_user_comm);

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_gen( FALSE, stderr, "^^^ **CM-being-recovered**: sent request for global recovery to CP: %i\n", assigned_cp_proc);
#endif

	MPI_Status s;
	MPI_Recv(NULL, 0, MPI_CHAR, MPI_ANY_SOURCE, CP_GLOBAL_RECOVERY_REQUEST, (*desc)->initial_user_comm, &s);
	assert(s.MPI_SOURCE == assigned_cp_proc);

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_gen( FALSE, stderr, "^^^ **CM-being-recovered**: received global recovery request from CP: %i\n", assigned_cp_proc);
#endif

	MPI_Send(NULL, 0, MPI_CHAR, assigned_cp_proc, CM_GLOBAL_RECOVERY_CONFIRMATION, (*desc)->initial_user_comm);

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_gen( FALSE, stderr, "^^^ **CM-being-recovered**: sent global recovery confirmation to CP: %i\n", assigned_cp_proc);
#endif

	NBC_Ibarrier(*(*desc)->flib_level_comm, &((*cm_desc)->global_recovery_request) );
	(*desc)->global_recovery_started = TRUE;

	uint64_t* checkpoint_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->data_offset];
	uint64_t* checkpoint_epoch_data = &((uint64_t*)(*desc)->cp_data)[(*desc)->epoch_data_offset];

	uint64_t* app_data = (uint64_t*) ((*((*cm_desc)->app_win))->win_array)[(*desc)->flib_level_rank].base;
	uint64_t* epoch_data = (uint64_t*) (*cm_desc)->epoch_data;

	int app_data_size = (*desc)->app_data_size;
	uint32_t epoch_data_size = (*desc)->epoch_data_size;
	int cp_data_unit = (*desc)->cp_data_disp_unit;

	MPI_Recv(checkpoint_data, app_data_size / cp_data_unit, MPI_UINT64_T, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_gen( FALSE, stderr, "^^^ **CM-being-recovered**: received checkpoint data from CP: %i\n", assigned_cp_proc);
#endif

	assert(s.MPI_SOURCE == assigned_cp_proc);
	assert(s.MPI_TAG == CP_RECOVERY_CHECKPOINT);

	memcpy(app_data, checkpoint_data, app_data_size);
	//memcpy(epoch_data, checkpoint_epoch_data, epoch_data_size);

	replay_actions();

	finish_global_recovery();

	(*desc)->global_recovery_started = FALSE;
}

void replay_actions() {
	//TODO: make it more efficient by adding some structures that will 'remember'
	//if we communicated or not with respective procs

#ifdef FLIB_RECOVERY_DEBUG
	debug_message_gen_e( FALSE, stderr, "^^^ **CM-being-recovered**: begin to replay my actions\n");
#endif

	char** logs = malloc(sizeof(char*) * (*desc)->cm_size);
	assert(logs != NULL);

	int* ptrs = malloc(sizeof(int) * (*desc)->cm_size);
	assert(ptrs != NULL);

	int* log_sizes = malloc(sizeof(int) * (*desc)->cm_size);
	assert(log_sizes != NULL);

	int* log_nodes = malloc(sizeof(int) * (*desc)->cm_size);
	assert(log_nodes != NULL);

	int* log_sizes_analyzed = calloc((*desc)->cm_size, sizeof(int));
	assert(log_sizes_analyzed != NULL);

	int* log_nodes_analyzed = calloc((*desc)->cm_size, sizeof(int));
	assert(log_nodes_analyzed != NULL);

	logs[(*desc)->flib_level_rank] = NULL;

#ifdef FLIB_RECOVERY_DEBUG
	debug_message_gen_e( FALSE, stderr, "^^^ **CM-being-recovered**: data structures initialized\n");
#endif

	for(int i = 0; i < (*desc)->cm_size; i++) {
		if(i == (*desc)->flib_level_rank) {
			ptrs[i] = -1;
			continue;
		}

		ptrs[i] = 0;

		//TODO: add checking the markers

		foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, i, 0, *(*cm_desc)->log_space->local_list);

		int nr_of_nodes_local = -1;
		int log_size = -1;

		foMPI_Get(&nr_of_nodes_local, 1, MPI_INT, i, 0, 1, MPI_INT, *(*cm_desc)->log_space->local_list_control);
		foMPI_Get(&log_size, 1, MPI_INT, i, 0, 1, MPI_INT, *(*cm_desc)->log_space->local_list_size);

		foMPI_Win_flush(i, *(*cm_desc)->log_space->local_list_control);
		foMPI_Win_flush(i, *(*cm_desc)->log_space->local_list_size);

		logs[i] = malloc(sizeof(char) * log_size);
		log_sizes[i] = log_size;
		log_nodes[i] = nr_of_nodes_local;

		foMPI_Get(&(logs[i][0]), log_size, MPI_CHAR, i, 0, log_size, MPI_CHAR, *(*cm_desc)->log_space->local_list);
		foMPI_Win_flush(i, *(*cm_desc)->log_space->local_list);

		foMPI_Win_unlock(i, *(*cm_desc)->log_space->local_list);
	}

#ifdef FLIB_RECOVERY_DEBUG
	debug_message_gen_e( FALSE, stderr, "^^^ **CM-being-recovered**: logs fetched\n");
#endif

	//get the minimum GNC
	//TODO: address the issue of max counter values
	int min_GNC = INT_MAX;
	
//	Common_params* params = malloc(sizeof(Common_params));

	for(int i = 0; i < (*desc)->cm_size; i++) {
		if(i == (*desc)->flib_level_rank || log_sizes[i] == 0) {
			continue;
		}

		Common_params params;
		memcpy(&params, logs[i], sizeof(Common_params));
		
		//debug_message_gen( FALSE, stderr, "^^^ **CM-being-recovered**: iter-p %i\n",i);

		//debug_message_gen( FALSE, stderr, "^^^ **CM-being-recovered**: logs: %c\n", logs[i][0]);

		if(params.GNC < min_GNC) {
			min_GNC = params.GNC;
		}
	}

#ifdef FLIB_RECOVERY_DEBUG
	debug_message_gen_e( FALSE, stderr, "^^^ **CM-being-recovered**: min GNC found\n");
#endif

	int recovery_done = FALSE;
	int analyzed_procs = 1;

	while(!recovery_done) {
		for(int i = 0; i < (*desc)->cm_size; i++) {
			if(i == (*desc)->flib_level_rank) {
				continue;
			}

			//TODO: improve the condition
			while(1) {
				Common_params params;
				memcpy(&params, &(logs[i][ptrs[i]]), sizeof(Common_params));
				ptrs[i] += sizeof(Common_params);
				//log_sizes_analyzed[i] += sizeof(Common_params);

				assert(params.GNC >= min_GNC);

				if(params.GNC > min_GNC) {
					break;
				}

				Put_get_args args;
				memcpy(&args, &(logs[i][ptrs[i]]), sizeof(Put_get_args));
				ptrs[i] += sizeof(Put_get_args);

				if(params.target_rank == (*desc)->flib_level_rank) {
					foMPI_Put(&(logs[i][ptrs[i]]), args.origin_count, args.origin_datatype, (*desc)->flib_level_rank, args.target_disp, args.target_count, args.target_datatype, *(*cm_desc)->app_win);
					foMPI_Win_flush((*desc)->flib_level_rank, *(*cm_desc)->app_win);
				}

				ptrs[i] += params.next_node_offset;
				log_sizes_analyzed[i] += params.next_node_offset;
				log_nodes_analyzed[i] += 1;

				if(log_nodes_analyzed[i] == log_nodes[i] || log_sizes_analyzed[i] == log_sizes[i]) {
					debug_message_gen( FALSE, stderr, "^^^ **CM-being-recovered**: log_sizes_analyzed: %i, log_sized: %i\n", log_sizes_analyzed[i], log_sizes[i]);
					assert(log_sizes_analyzed[i] == log_sizes[i]);
					assert(log_nodes_analyzed[i] == log_nodes[i]);
					analyzed_procs++;
					break;
				}
			}
		}
		min_GNC++;

#ifdef FLIB_RECOVERY_DEBUG
	debug_message_gen( FALSE, stderr, "^^^ **CM-being-recovered**: logs for GNC %i replayed\n", min_GNC);
#endif

		if(analyzed_procs == (*desc)->cm_size) {
			recovery_done = TRUE;
		}
	}

	for(int i = 0; i < (*desc)->cm_size; i++) {
		if(i != (*desc)->flib_level_rank) {
			free(logs[i]);
		}
	}

	free(ptrs);
	free(log_sizes);
	free(log_nodes);
	free(log_sizes_analyzed);
	free(log_nodes_analyzed);
}

//#######################################################################################################################################
//A routine that finishes the recovery
//#######################################################################################################################################

void finish_global_recovery() {
#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>>>>>>>>>>>>> GLOBAL RECOVERY: begin to synchronize <<<<<<<<<<<<<\n");
#endif

	MPI_Barrier((*desc)->initial_user_comm);

#ifdef FLIB_GLOBAL_RECOVERY_DEBUG
	debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, ">>>>>>>>>>>>>> GLOBAL RECOVERY PERFORMED <<<<<<<<<<<<<\n");
#endif
}

//#######################################################################################################################################
//TODO: a routine that integrates the local/global routines in one
//#######################################################################################################################################

void recover_me() {
	if(do_logging()) {
		cm_local_recovery();
	}
	else {
		cm_global_recovery();
	}
}

//#######################################################################################################################################
//TODO: routines that address other recovery schemes that may require self-locking
//#######################################################################################################################################

void lock_for_recovery() {
	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(*cm_desc)->app_win);
	//foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(*cm_desc)->log_space->puts_win);
}

void unlock_recovery() {
	foMPI_Win_unlock((*desc)->flib_level_rank, *(*cm_desc)->app_win);
	//foMPI_Win_unlock((*desc)->flib_level_rank, *(*cm_desc)->log_space->puts_win);
}

//#######################################################################################################################################
//A routine for a recovery that does NOT involve active participation of other procs
//#######################################################################################################################################

void cm_local_recovery() {
}


