#include "flib.h"
#include "flib_groups.h"
#include "flib_tests.h"
#include "flib_debug.h"
#include "flib_cm_proc.h"
#include "flib_recovery.h"

int __testing_puts_nr = 0;
int* __testing_gets_nr = NULL;
__Op_node** __local_log_array = NULL;
gdsl_queue_t* __global_log_array = NULL;

//TODO: do more extensive testing
void __test_groups_division() {
	MPI_Barrier((*desc)->initial_user_comm);

	if(cp_proc()) {
		int cp_procs = -1;
		MPI_Comm_size(*(*desc)->flib_level_comm, &cp_procs);
		int assigned_procs_nr = assigned_cm_procs_nr();

		assert(assigned_procs_nr * cp_procs + cp_procs == get_initial_user_size());

		int i = 0;
		MPI_Status s;
		for(; i < assigned_procs_nr; i++) {
			MPI_Recv(NULL, 0, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, (*desc)->initial_user_comm, &s);

			assert( s.MPI_TAG == DEBUG_CM_INITIAL_CONFIRMATION);
			assert(is_our_proc( s.MPI_SOURCE, _FLIB_INITIAL_COMM ) == TRUE);
		}
	}
	else {
		int my_cp_proc = my_assigned_cp_proc();
		MPI_Send(NULL, 0, MPI_CHAR, my_cp_proc, DEBUG_CM_INITIAL_CONFIRMATION, (*desc)->initial_user_comm);

		int calculated_world_rank = translate_from_flib_to_world_rank((*desc)->flib_level_rank);

		assert(calculated_world_rank == (*desc)->initial_user_rank);
	}

	MPI_Barrier((*desc)->initial_user_comm);
}


void __test_initial_checkpoint() {
	MPI_Barrier((*desc)->initial_user_comm);

	if(cp_proc()) {
		int app_data_size = (*desc)->app_data_size;
		int epoch_data_size = (*desc)->epoch_data_size;	
		int data_size = app_data_size + epoch_data_size;	
		int cp_data_unit = (*desc)->cp_data_disp_unit;

		int xor_offset = (*desc)->xor_data_offset;
		uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[xor_offset];

		uint64_t* remote_data = (uint64_t*) malloc( data_size * sizeof(char) );
		uint64_t* xor_result = (uint64_t*) malloc( data_size * sizeof(char) );

		int* assigned_procs = assigned_cm_procs();

		int j = 0;
		for(; j < assigned_cm_procs_nr(); j++) {

			foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, assigned_procs[j], 0, *(*desc)->cp_win);

			foMPI_Get(remote_data, data_size / cp_data_unit, MPI_UINT64_T, assigned_procs[j], 0, data_size / cp_data_unit, MPI_UINT64_T, *(*desc)->cp_win);
			foMPI_Win_flush(assigned_procs[j], *(*desc)->cp_win);

			foMPI_Win_unlock(assigned_procs[j], *(*desc)->cp_win);

			if(j == 0) {
				memcpy(xor_result, remote_data, data_size);
			}
			else {
				calculate_xor(remote_data, xor_result, data_size);
			}

		}

		int w = 0;
		for(; w < data_size / cp_data_unit; w++) {
			assert(xor_result[w] == xored_data[w]);
		}

		// we confirmed that One sided operations generate the same result as P2P.
		// now let's examine whether we can generate remaining CM procs using XORING

		uint64_t* temp = (uint64_t*) malloc( data_size * sizeof(char) );
		memset(temp, 0, data_size);
		memset(remote_data, 0, data_size);
		memset(xor_result, 0, data_size);

		j = 0;
		for(; j < assigned_cm_procs_nr(); j++) {
			memcpy(xor_result, xored_data, data_size);

			foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, assigned_procs[j], 0, *(*desc)->cp_win);
			foMPI_Get(remote_data, data_size / cp_data_unit, MPI_UINT64_T, assigned_procs[j], 0, data_size / cp_data_unit, MPI_UINT64_T, *(*desc)->cp_win);
			foMPI_Win_flush(assigned_procs[j], *(*desc)->cp_win);
			foMPI_Win_unlock(assigned_procs[j], *(*desc)->cp_win);

			int k = 0;
			for(; k < assigned_cm_procs_nr(); k++) {
				if(k == j) {
					continue;
				}

				foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, assigned_procs[k], 0, *(*desc)->cp_win);
				foMPI_Get(temp, data_size / cp_data_unit, MPI_UINT64_T, assigned_procs[k], 0, data_size / cp_data_unit, MPI_UINT64_T, *(*desc)->cp_win);
				foMPI_Win_flush(assigned_procs[k], *(*desc)->cp_win);
				foMPI_Win_unlock(assigned_procs[k], *(*desc)->cp_win);

				calculate_xor(temp, xor_result, data_size);
			}

			int w = 0;
			for(; w < data_size / cp_data_unit; w++) {
				assert(xor_result[w] == remote_data[w]);
			}
		}

		fprintf(stderr,"[%i] CP process: my initial checkpoint XOR data is correct\n", (*desc)->initial_user_rank);

		free(remote_data);
		free(xor_result);
		free(temp);	
		free(assigned_procs);
	}
	else {
	}	

	MPI_Barrier((*desc)->initial_user_comm);

	if((*desc)->initial_user_rank == 0) {
		fprintf(stderr,"=============== initial CP testing done =============== \n");
		fflush(stderr);
	}

	MPI_Barrier((*desc)->initial_user_comm);
}

//TODO: finish
void __test_checkpoint(int origin_rank) {
	int app_data_size = (*desc)->app_data_size;
	int epoch_data_size = (*desc)->epoch_data_size;  
	int data_size = app_data_size + epoch_data_size;
	int cp_data_unit = (*desc)->cp_data_disp_unit;

	int xor_offset = (*desc)->xor_data_offset;
	uint64_t* xored_data = &((uint64_t*)(*desc)->cp_data)[xor_offset];

	uint64_t* result = (uint64_t*) malloc( data_size * sizeof(char) );
	uint64_t* remote_data = (uint64_t*) malloc( data_size * sizeof(char) );

	//TODO: probably unnecessary, only CP process runs this code anyway in the current testing version
	if(cp_proc()) {

		int* assigned_procs = assigned_cm_procs();

#ifdef FLIB_VERBOSE_DEBUG
		fprintf(stderr, "[%i] CP: begin to TEST the xoring results...\n", (*desc)->initial_user_rank);
#endif

		memcpy(result, xored_data, data_size);

		int j = assigned_procs[0];
		for(; j <= assigned_procs[1]; j++) {

			if(j == origin_rank) {
				continue;
			} 

			foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, j, 0, *(*desc)->cp_win);

			foMPI_Get(remote_data, data_size / cp_data_unit, MPI_UINT64_T, j, 0, data_size / cp_data_unit, MPI_UINT64_T, *(*desc)->cp_win);
			foMPI_Win_flush(j, *(*desc)->cp_win);

			foMPI_Win_unlock(j, *(*desc)->cp_win);

			calculate_xor(remote_data, result, data_size);
		}

		foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, origin_rank, 0, *(*desc)->cp_win);

		foMPI_Get(remote_data, data_size / cp_data_unit, MPI_UINT64_T, origin_rank, 0, data_size / cp_data_unit, MPI_UINT64_T, *(*desc)->cp_win);
		foMPI_Win_flush(origin_rank, *(*desc)->cp_win);

		foMPI_Win_unlock(origin_rank, *(*desc)->cp_win);

#ifdef FLIB_VERBOSE_DEBUG

		//fprintf(stderr, "[%i] CP: TEST xoring done, sending the results to CM %i. XOR data: %c %c\n", (*desc)->initial_user_rank, origin_rank, result[0], result[1]);

		fprintf(stderr, "[%i] CP: TEST: xoring calculation done, final XOR data: %f %f. Remote data (from %i): %f %f\n", (*desc)->initial_user_rank,
				((double*)result)[0], ((double*)result)[1], origin_rank, ((double*)remote_data)[0], ((double*)remote_data)[1]);

#endif

		int v = 0;
		for(; v < data_size / cp_data_unit; v++) {
			assert(result[v] == remote_data[v]);
		}

		//MPI_Send(result, data_size, MPI_CHAR, origin_rank, DEBUG_CP_CHECKPOINT_TEST, (*desc)->initial_user_comm);

#ifdef FLIB_DEBUG

		//fprintf(stderr, "[%i] CP: TEST xoring done, sending the results to CM %i. XOR data: %c %c\n", (*desc)->initial_user_rank, origin_rank, result[0], result[1]);

		fprintf(stderr, "[%i] CP: TEST xoring done, checksum OK\n", (*desc)->initial_user_rank);

#endif

		free(assigned_procs);
	}
	else {

	}

	free(result);
	free(remote_data);
}

void __init_logging_testing() {
	__init_puts_logging();
	__init_gets_logging();
}

void __init_puts_logging() {
	__local_log_array = malloc( sizeof(__Op_node*) * MAX_LOCAL_LOG_SIZE );
	assert(__local_log_array != NULL);

	int i = 0;
	for(i = 0; i < MAX_LOCAL_LOG_SIZE; i++) {
		__local_log_array[i] = NULL;
	}
	__testing_puts_nr = 0;
}

void __init_gets_logging() {
	__global_log_array = malloc( sizeof(gdsl_queue_t) * (*desc)->cm_size );
	assert(__global_log_array != NULL);

	__testing_gets_nr = malloc( sizeof(int) * (*desc)->cm_size );
	assert(__testing_gets_nr != NULL);

	int i = 0;
	for(i = 0; i < (*desc)->cm_size; i++) {
		__global_log_array[i] = gdsl_queue_alloc(NULL, NULL, NULL);
		__testing_gets_nr[i] = 0;
	}
}

void __log_put( const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype, MPI_Op operation,
		Operation_type op_type, int verbose) {
	if(verbose == TRUE) {
		int left = MAX_LOCAL_LOG_SIZE - __testing_puts_nr;
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: begin to log a PUT (%lld). Number of puts logged: %i, left: %i\n", *( (int64_t*)(origin_addr) ), __testing_puts_nr, left);
	}

	__Op_node* node = malloc( sizeof(__Op_node) );
	__local_log_array[__testing_puts_nr] = node;
	__testing_puts_nr++;

	if(__testing_puts_nr >= MAX_LOCAL_LOG_SIZE) {
		free(node);

		int j = 0;
		for(; j < __testing_puts_nr - 1; j++) {

			free(__local_log_array[j]->payload);
			free(__local_log_array[j]);
		}

		debug_message_gen_e( TRUE, stderr, "CM: ERROR: TEST LOGGING: NO MEMORY LEFT!!!!! Exiting...\n");
	}

	node->op_type = op_type;

	int epoch_nr = (*cm_desc)->local_epoch_counter[target_rank];
	node->epoch_nr = epoch_nr;

	node->origin_count = origin_count;
	node->origin_datatype = origin_datatype;
	node->target_rank = target_rank;
	node->target_disp = target_disp;
	node->target_count = target_count;
	node->target_datatype = target_datatype;
	node->operation = operation;

	int type_size = -1;
	MPI_Type_size(origin_datatype, &type_size);

	int payload_size = origin_count * type_size;
	node->payload = malloc( payload_size * sizeof(char) );

	memcpy(node->payload, origin_addr, payload_size);

	if(verbose) {
		int left = MAX_LOCAL_LOG_SIZE - __testing_puts_nr;
		Log_space* log_space = (*cm_desc)->log_space;
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: PUT logged (%lld - %lld). Number of puts logged: %i, left: %i\n",
				*( (int64_t*)(origin_addr) ), *( (int64_t*)(node->payload) ),  __testing_puts_nr, left);
		if(__testing_puts_nr != log_space->number_of_nodes_local) {
			debug_message_gen( TRUE, stderr, " CM: : PUT LOGGED: ERROR: __testing_puts_nr(%i) != log_space->number_of_nodes_local(%i)\n", __testing_puts_nr, log_space->number_of_nodes_local);
		}
		else {
			debug_message_gen( FALSE, stderr, " CM: : PUT LOGGED: OK: __testing_puts_nr(%i) == log_space->number_of_nodes_local(%i)\n", __testing_puts_nr, log_space->number_of_nodes_local);
		}
	}
}

void __log_get( const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype, int epoch_nr, int verbose) {
	if(verbose) {
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: begin to log a GET (%lld). Number of gets logged by me at this target: %i\n",
				*( (int64_t*)(origin_addr) ), __testing_gets_nr[target_rank]);
	}

	__Op_node* node = malloc( sizeof(__Op_node) );
	assert(node != NULL);

	node->op_type = GET;
	node->origin_count = origin_count;
	node->origin_datatype = origin_datatype;
	node->target_rank = target_rank;
	node->target_disp = target_disp;
	node->target_count = target_count;
	node->target_datatype = target_datatype;
	node->epoch_nr = epoch_nr;

	int type_size = -1;
	MPI_Type_size(origin_datatype, &type_size);

	int payload_size = origin_count * type_size;
	node->payload = malloc( payload_size * sizeof(char) );

	memcpy(node->payload, origin_addr, payload_size);

	gdsl_queue_insert(__global_log_array[target_rank], node);
	__testing_gets_nr[target_rank]++;

	if(verbose) {
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: GET logged (%lld - %lld). Number of gets logged for target rank %i: %i\n",
				*( (int64_t*)(origin_addr) ), *( (int64_t*)(node->payload) ),  target_rank, __testing_gets_nr[target_rank]);
	}
}

void __log_accumulate( const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype, MPI_Op operation, int verbose) {

	if(verbose) {
		int left = MAX_LOCAL_LOG_SIZE - __testing_puts_nr;
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: begin to log a ACCUMULATE. Number of puts logged: %i, left: %i\n", __testing_puts_nr, left);
	}

	__Op_node* node = malloc( sizeof(__Op_node) );
	__local_log_array[__testing_puts_nr] = node;
	__testing_puts_nr++;

	if(__testing_puts_nr >= MAX_LOCAL_LOG_SIZE) {
		free(node);

		int j = 0;
		for(; j < __testing_puts_nr - 1; j++) {

			free(__local_log_array[j]->payload);
			free(__local_log_array[j]);
		}

		debug_message_gen_e( TRUE, stderr, "CM: ERROR: TEST LOGGING: NO MEMORY LEFT!!!!! Exiting...\n");
	}

	node->op_type = ACCUMULATE;

	int epoch_nr = (*cm_desc)->local_epoch_counter[target_rank];

	node->origin_count = origin_count;
	node->origin_datatype = origin_datatype;
	node->target_rank = target_rank;
	node->target_disp = target_disp;
	node->target_count = target_count;
	node->target_datatype = target_datatype;
	node->operation = operation;

	node->epoch_nr = epoch_nr;

	int type_size = -1;
	MPI_Type_size(origin_datatype, &type_size);

	int payload_size = origin_count * type_size;
	node->payload = malloc( payload_size * sizeof(char) );

	memcpy(node->payload, origin_addr, payload_size);

	if(verbose == TRUE) {
		int left = MAX_LOCAL_LOG_SIZE - __testing_puts_nr;
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: ACCUMULATE logged. Number of puts logged: %i, left: %i\n", __testing_puts_nr, left);
	}
}

//TODO: probably not safe, test it
void __print_gets() {
	Log_space* log_space = (*cm_desc)->log_space;

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_list));
	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_logs_control_win));

	char* target_ptr = (*cm_desc)->log_space->global_list_data;
	int analyzed_bytes = 0;

	int64_t bytes_written = -1;
	int64_t faked_var = 0;
	foMPI_Get_accumulate(&faked_var, 1, MPI_INT64_T, &bytes_written, 1, MPI_INT64_T, (*desc)->flib_level_rank, (*desc)->cm_size, 1, MPI_INT64_T, foMPI_NO_OP,
			*(log_space->global_logs_control_win));
	foMPI_Win_flush((*desc)->flib_level_rank, *(log_space->global_logs_control_win));

	while(analyzed_bytes < bytes_written) {
		Common_params params;

		memcpy(&params, target_ptr, sizeof(Common_params));
		target_ptr += sizeof(Common_params);
		analyzed_bytes += sizeof(Common_params);

		fprintf(stderr,"[flib: %i] ================ NEW GET ================ \n", (*desc)->flib_level_rank);

		fprintf(stderr,"[flib: %i] origin_rank: %i\n", (*desc)->flib_level_rank, params.target_rank);
		fprintf(stderr,"[flib: %i] epoch nr: %i\n", (*desc)->flib_level_rank, params.epoch_nr);
		fprintf(stderr,"[flib: %i] op type: %i\n", (*desc)->flib_level_rank, params.op_type);

		Put_get_args args;

		memcpy(&args, target_ptr, sizeof(Put_get_args));
		target_ptr += sizeof(Put_get_args);
		analyzed_bytes += sizeof(Put_get_args);

		fprintf(stderr,"[flib: %i] origin count: %i\n", (*desc)->flib_level_rank, args.origin_count);
		fprintf(stderr,"[flib: %i] target count: %i\n", (*desc)->flib_level_rank, args.target_count);
		fprintf(stderr,"[flib: %i] target disp: %i\n", (*desc)->flib_level_rank, args.target_disp);

		int origin_type_size = -1;
		MPI_Type_size(args.origin_datatype, &origin_type_size);
		assert(origin_type_size > 0);

		int payload_size = args.origin_count * origin_type_size;

		void* payload = malloc( payload_size );

		memcpy(payload, target_ptr, payload_size );
		target_ptr += payload_size;
		analyzed_bytes += payload_size;

		fprintf(stderr,"[flib: %i] payload first part: %f\n", (*desc)->flib_level_rank, *((double*)payload));
	} 

	foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_logs_control_win));
	foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_list));
}


void __test_all_logs_of_gets(int verbose) {

	MPI_Barrier(*(*desc)->flib_level_comm);

	int __final_sum = -1;
	int __global_logs_del = 0;
	int __global_test_logs_del = 0;

	while(__final_sum != 0) {
		MPI_Allreduce(&((*cm_desc)->log_space->__cleared_global_logs), &__global_logs_del, 1, MPI_INT, MPI_SUM, *(*desc)->flib_level_comm);
		MPI_Allreduce(&((*cm_desc)->log_space->__cleared_global_test_logs), &__global_test_logs_del, 1, MPI_INT, MPI_SUM, *(*desc)->flib_level_comm);

		__final_sum = __global_logs_del + __global_test_logs_del;

		cm_process_notifications();
	}

	if(verbose == TRUE) {
		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, " CM: TEST LOGGING: START TESTING GETS!!!!.\n");
	}

	Log_space* log_space = (*cm_desc)->log_space;

	int* nr_of_logs = (int*) malloc(sizeof(int)*(*desc)->cm_size);
	for(int i = 0; i < (*desc)->cm_size; i++) {
		nr_of_logs[i] = 0;
	}

	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_list));
	foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, (*desc)->flib_level_rank, 0, *(log_space->global_logs_control_win));

	foMPI_Win_sync(*(log_space->global_list));
	foMPI_Win_sync(*(log_space->global_logs_control_win));

	//TEST Phase 1: we get the number of gets logged in our LG
	//recorded by different processes
	//then we communicate these numbers and compare
	//with local numbers of of logged gets

	char* target_ptr = (char*)(log_space->global_list_data);
	int next_free_offset = (log_space->global_logs_control_data)[(*desc)->cm_size];
	int nr_of_logged_gets = (log_space->global_logs_control_data)[(*desc)->cm_size + 1];
	int nodes_analyzed = 0;
	int analyzed_bytes = 0;

	Common_params params;

	while(nodes_analyzed < nr_of_logged_gets) {
		memcpy(&params, target_ptr, sizeof(Common_params));
		target_ptr += sizeof(Common_params);
		analyzed_bytes += sizeof(Common_params);

		//TODO: this name (target rank) is misleading
		//it relates to the origin actually
		assert(params.target_rank >= 0 && params.target_rank < (*desc)->cm_size);
		nr_of_logs[params.target_rank]++;

		assert(params.epoch_nr >= 0);

		//TODO: make it more generic if useful
		assert(params.op_type == GET);

		Put_get_args args;

		memcpy(&args, target_ptr, sizeof(Put_get_args));
		target_ptr += sizeof(Put_get_args);
		analyzed_bytes += sizeof(Put_get_args);

		assert(args.origin_count >= 0);
		assert(args.target_disp  >= 0);
		assert(args.target_count >= 0);

		int origin_type_size = -1;
		MPI_Type_size(args.origin_datatype, &origin_type_size);
		assert(origin_type_size > 0);

		int payload_size = args.origin_count * origin_type_size;

		char* payload = (char*) malloc( payload_size );

		memcpy(payload, target_ptr, payload_size );
		target_ptr += payload_size;
		analyzed_bytes += payload_size;

		free(payload);
		nodes_analyzed++;
	}

	assert(next_free_offset == analyzed_bytes);

	MPI_Barrier(*(*desc)->flib_level_comm);

	int* my_logs_at_remote_procs = (int*) calloc((*desc)->cm_size,sizeof(int));
	//MPI_Alltoall(&nr_of_logs[0], (*desc)->cm_size, MPI_INT, &my_logs_at_remote_procs[0], (*desc)->cm_size, MPI_INT, *(*desc)->flib_level_comm);

	for(int i = 0; i < (*desc)->cm_size; i++) {
		if(i == (*desc)->flib_level_rank) {
			continue;
		}
		assert(nr_of_logs[i] >= 0);
		MPI_Send(&(nr_of_logs[i]), 1, MPI_INT, i, 111, *(*desc)->flib_level_comm);
	}

	int* nr = (int*)malloc(sizeof(int));
	for(int i = 0; i < (*desc)->cm_size; i++) {
		if(i == (*desc)->flib_level_rank) {
			continue;
		}
		MPI_Status s;
		MPI_Recv(nr, 1, MPI_INT, MPI_ANY_SOURCE, 111, *(*desc)->flib_level_comm, &s);
		my_logs_at_remote_procs[s.MPI_SOURCE] += *nr;
	}
	free(nr);

	for(int i = 0; i < (*desc)->cm_size; i++) {
		if(my_logs_at_remote_procs[i] != __testing_gets_nr[i]) {
			debug_message_gen( TRUE, stderr, " CM: ERROR: remote_logs for proc [%i,CM]: %i != __testing_logs: %i\n", i, my_logs_at_remote_procs[i], __testing_gets_nr[i]);
		}
		else {
			if(my_logs_at_remote_procs[i] != 0) {
				debug_message_gen( FALSE, stderr, " CM: remote_logs for proc [%i,CM]: %i == __testing_logs: %i\n", i, my_logs_at_remote_procs[i], __testing_gets_nr[i]);
			}
		}
		assert(my_logs_at_remote_procs[i] == __testing_gets_nr[i]);
		assert(my_logs_at_remote_procs[i] == gdsl_queue_get_size(__global_log_array[i]));
	}

	int sum_get_logs = 0;
	for(int i = 0; i < (*desc)->cm_size; i++) {
		sum_get_logs += __testing_gets_nr[i];
	}

	if(verbose == TRUE) {
		debug_message_gen_e( FALSE, stderr, " CM: TESTING LOGGED GETS (PHASE 1) finished CORRECTLY.\n");
	}

	MPI_Barrier(*(*desc)->flib_level_comm);

	//PHASE 2: we sent logs one by one
	//and compare all fields with the local logs
	//making sure there is no data corruption

	target_ptr = (char*)(log_space->global_list_data);
	next_free_offset = (log_space->global_logs_control_data)[(*desc)->cm_size];
	nr_of_logged_gets = (log_space->global_logs_control_data)[(*desc)->cm_size + 1];
	nodes_analyzed = 0;
	analyzed_bytes = 0;

	while(nodes_analyzed < nr_of_logged_gets) {
		memcpy(&params, target_ptr, sizeof(Common_params));
		target_ptr += sizeof(Common_params);
		analyzed_bytes += sizeof(Common_params);

		//TODO: this name (target rank) is misleading
		//it relates to the origin actually
		assert(params.target_rank >= 0 && params.target_rank < (*desc)->cm_size);
		assert(params.epoch_nr >= 0);
		assert(params.op_type == GET); //TODO: more generic

		Put_get_args args;

		memcpy(&args, target_ptr, sizeof(Put_get_args));
		target_ptr += sizeof(Put_get_args);
		analyzed_bytes += sizeof(Put_get_args);

		assert(args.origin_count >= 0);
		assert(args.target_disp  >= 0);
		assert(args.target_count >= 0);

		int origin_type_size = -1;
		MPI_Type_size(args.origin_datatype, &origin_type_size);
		assert(origin_type_size > 0);

		int payload_size = args.origin_count * origin_type_size;

		char* log_to_sent = (char*) malloc( (sizeof(params) + sizeof(args) + payload_size) * sizeof(char));
		memcpy(&log_to_sent[0], &params, sizeof(params));
		memcpy(&log_to_sent[sizeof(params)], &args, sizeof(args));
		memcpy(&log_to_sent[sizeof(params)+sizeof(args)], target_ptr, payload_size);

		int size_to_send = sizeof(Common_params) + sizeof(Put_get_args) + payload_size;

		MPI_Send(&log_to_sent[0], size_to_send, MPI_CHAR, params.target_rank, 112, *(*desc)->flib_level_comm);

		target_ptr += payload_size;
		analyzed_bytes += payload_size;

		nodes_analyzed++;
		free(log_to_sent);
	}

	MPI_Status s;
	int size;
	int* received_logs = (int*) calloc((*desc)->cm_size, sizeof(int));
	for(int i = 0; i < sum_get_logs; i++) {
		MPI_Probe(MPI_ANY_SOURCE, 112, *(*desc)->flib_level_comm, &s);
		MPI_Get_count(&s, MPI_CHAR, &size);
		char* buf = (char*) malloc(size * sizeof(char));
		MPI_Recv(&buf[0], size, MPI_CHAR, s.MPI_SOURCE, 112, *(*desc)->flib_level_comm, &s);

		Common_params params;
		Put_get_args args;
		int payload_size = size - (sizeof(Common_params) + sizeof(Put_get_args));
		char* payload = (char*) malloc(payload_size * sizeof(char));

		memcpy(&params, &buf[0], sizeof(Common_params));
		memcpy(&args, &buf[sizeof(Common_params)], sizeof(Put_get_args));
		memcpy(&payload[0], &buf[sizeof(Common_params)+sizeof(Put_get_args)], payload_size);

		int queue_size =  gdsl_queue_get_size(__global_log_array[s.MPI_SOURCE]);
		__Op_node* node = gdsl_queue_search_by_position(__global_log_array[s.MPI_SOURCE], queue_size - received_logs[s.MPI_SOURCE]);

		if(node->epoch_nr != params.epoch_nr) {
			debug_message_gen( FALSE, stderr, " params.epoch_nr: %i != node->epoch_nr: %i.\n", params.epoch_nr, node->epoch_nr);
			assert(node->epoch_nr == params.epoch_nr);
		}
		assert(node->target_rank == s.MPI_SOURCE);

		assert(args.origin_count == node->origin_count);
		assert(args.origin_datatype == node->origin_datatype);
		assert(args.target_disp == node->target_disp);
		assert(args.target_count == node->target_count);
		assert(args.target_datatype == node->target_datatype);

		for(int j = 0; j < payload_size; j++) {
			assert(payload[j] == (node->payload)[j]);
		}

		received_logs[s.MPI_SOURCE]++;

		free(buf);
		free(payload);
	}

	free(nr_of_logs);
	free(my_logs_at_remote_procs);
	free(received_logs);

	if(verbose == TRUE) {
		debug_message_gen_e( FALSE, stderr, " CM: TESTING LOGGED GETS (PHASE 2) finished CORRECTLY.\n");
	}

	foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_list));
	foMPI_Win_unlock((*desc)->flib_level_rank, *(log_space->global_logs_control_win));

	MPI_Barrier(*(*desc)->flib_level_comm);
}




/* commenting out this function
	 probably won't be used 
	 but may be actually useful...
	 void __test_consistency_of_gets(int verbose) {
	 if(verbose == TRUE) {
	 debug_message_gen_e( FALSE, stderr, " CM: TEST LOGGING: START TESTING LOCAL GETS.\n");
	 }
	 Log_space* log_space = (*cm_desc)->log_space;

	 char* target_ptr = NULL;
	 char* payload = NULL;

	 char* remote_global_list_data = NULL;
	 int64_t remote_next_free_offset = -1;

	 int i = 0;
	 __Op_node* node = NULL;

	 for(; i < (*desc)->cm_size; i++) {
	 if( (log_space->procs_communicated_using_gets)[i] == TRUE ) {
	 if(verbose == TRUE) {
	 debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: START TESTING GETS from rank %i. Number of gets logged there: %i\n", __testing_gets_nr[i]);
	 }
	 int64_t remote_sum = -1;

	 foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, i, 0, *(log_space->global_list));
	 foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, i, 0, *(log_space->global_logs_control_win));

	 foMPI_Get(&remote_sum, 1, MPI_INT64_T, i, (*desc)->flib_level_rank, 1, MPI_INT64_T, *(log_space->global_logs_control_win));
	 foMPI_Win_flush(i, *(log_space->global_logs_control_win));

	 assert(remote_sum != 0);
	 assert(remote_sum > 0);

	 foMPI_Get(&remote_next_free_offset, 1, MPI_INT64_T, i, (*desc)->cm_size, 1, MPI_INT64_T, *(log_space->global_logs_control_win));
	 foMPI_Win_flush(i, *(log_space->global_logs_control_win));

	 assert(remote_next_free_offset > 0);

	 remote_global_list_data = malloc( sizeof(char) * remote_next_free_offset );
	 assert(remote_global_list_data != NULL);

	 foMPI_Get(remote_global_list_data, (int)remote_next_free_offset, MPI_CHAR, i, 0, (int)remote_next_free_offset, MPI_CHAR, *(log_space->global_list));
	 foMPI_Win_flush(i, *(log_space->global_list));

	 target_ptr = &remote_global_list_data[0];

	 int analyzed_bytes = 0;
	 int test_nodes_compared = 0;
	 int remote_nodes_compared = 0;
	 int j = 0;
	 for(; j < __testing_gets_nr[i]; j++) {
	 node =  ((__Op_node*)(gdsl_queue_search_by_position(__global_log_array[i],j)));

	 assert(node->target_rank == i);
	 test_nodes_compared++;
	 Common_params params;

	 do {
	 memcpy(&params, target_ptr, sizeof(Common_params));
	 target_ptr += sizeof(Common_params);
	 analyzed_bytes += sizeof(Common_params);

	 if(params.target_rank == (*desc)->flib_level_rank) {
	 remote_nodes_compared++;

	 assert(params.epoch_nr == node->epoch_nr);
	 assert(params.op_type == node->op_type);

//fprintf(stderr,"[%i] YEA\n", (*desc)->flib_level_rank);			
// TODO: for now
assert(params.op_type == GET);

Put_get_args args;

memcpy(&args, target_ptr, sizeof(Put_get_args));
target_ptr += sizeof(Put_get_args);
analyzed_bytes += sizeof(Put_get_args);

assert(args.origin_count == node->origin_count);
assert(args.origin_datatype == node->origin_datatype);
assert(args.target_disp == node->target_disp);
assert(args.target_count == node->target_count);
assert(args.target_datatype == node->target_datatype);

int origin_type_size = -1;
MPI_Type_size(args.origin_datatype, &origin_type_size);
assert(origin_type_size > 0);

int payload_size = args.origin_count * origin_type_size;

payload = (char*) malloc( payload_size );

memcpy(payload, target_ptr, payload_size );
target_ptr += payload_size;
analyzed_bytes += payload_size;

int w = 0;
for(; w < payload_size; w++) {
	assert(payload[w] == node->payload[w]);
}

free(payload);

goto next_test_node_loop;
}
else {
	Put_get_args args;

	memcpy(&args, target_ptr, sizeof(Put_get_args));
	target_ptr += sizeof(Put_get_args);
	analyzed_bytes += sizeof(Put_get_args);

	int origin_type_size = -1;
	MPI_Type_size(args.origin_datatype, &origin_type_size);
	assert(origin_type_size > 0);

	int payload_size = args.origin_count * origin_type_size;

	target_ptr += payload_size;
	analyzed_bytes += payload_size;
}		
} while(analyzed_bytes < remote_next_free_offset - 1);

next_test_node_loop:;
}

assert(test_nodes_compared == remote_nodes_compared);

//fprintf(stderr,"[%i] DDD\n", (*desc)->initial_user_rank);

free(remote_global_list_data);

foMPI_Win_unlock(i, *(log_space->global_list));
foMPI_Win_unlock(i, *(log_space->global_logs_control_win));
}
}
if(verbose == TRUE) {
	debug_message_gen_e( FALSE, stderr, " CM: TEST LOGGING: TESTING LOCAL >>GETS<< finished CORRECTLY.\n");
}
}
*/

void __test_consistency_of_puts(int verbose) {
	Log_space* log_space = (*cm_desc)->log_space;

	char* target_ptr = ((char*)(log_space->local_list_data));
	char* payload = NULL;

	int i = 0;
	__Op_node* node = NULL;

	if(verbose == TRUE) {
		int left = MAX_LOCAL_LOG_SIZE - __testing_puts_nr;
		debug_message_gen( FALSE, stderr, " CM: START TESTING LOCAL LOGS. Number of puts logged: %i, left: %i\n", __testing_puts_nr, left);
		if(__testing_puts_nr != log_space->number_of_nodes_local) {
			debug_message_gen( TRUE, stderr, " CM: : TESTING: ERROR: __testing_puts_nr(%i) != log_space->number_of_nodes_local(%i)\n", __testing_puts_nr, log_space->number_of_nodes_local);
		}
		else {
			debug_message_gen( FALSE, stderr, " CM: : TESTING: OK: __testing_puts_nr(%i) == log_space->number_of_nodes_local(%i)\n", __testing_puts_nr, log_space->number_of_nodes_local);
		}
	}

	assert(__testing_puts_nr == log_space->number_of_nodes_local);

	for(; i < __testing_puts_nr; i++) {
		node = __local_log_array[i];

		Common_params params;

		memcpy(&params, target_ptr, sizeof(Common_params));
		target_ptr += sizeof(Common_params);

		//if(node->epoch_nr != params.epoch_nr) fprintf(stderr,"[%i] test epoch: %i, logged epoch: %i\n", (*desc)->initial_user_rank, node->epoch_nr, params.epoch_nr);

		assert(params.epoch_nr == node->epoch_nr);
		assert(params.target_rank == node->target_rank);
		assert(params.op_type == node->op_type);

		if(params.op_type == PUT) {
			Put_get_args args;

			memcpy(&args, target_ptr, sizeof(Put_get_args));
			target_ptr += sizeof(Put_get_args);

			assert(args.origin_count == node->origin_count);
			assert(args.origin_datatype == node->origin_datatype);
			assert(args.target_disp == node->target_disp);
			assert(args.target_count == node->target_count);
			assert(args.target_datatype == node->target_datatype);

			int origin_type_size = -1;
			MPI_Type_size(args.origin_datatype, &origin_type_size);
			int payload_size = args.origin_count * origin_type_size;

			payload = (char*) malloc( payload_size );

			memcpy(payload, target_ptr, payload_size );
			target_ptr += payload_size;

			int j = 0;
			for(; j < payload_size; j++) {
				if(payload[j] != node->payload[j]) {
					fprintf(stderr,"[%i] ERROR: FUCK, original: %c, tested: %c, (%lld - %lld)\n", 
							(*desc)->initial_user_rank, payload[j], node->payload[j], *( (int64_t*)(payload) ), *( (int64_t*)(node->payload) ));
				}

				assert(payload[j] == node->payload[j]);
				//fprintf(stderr,"[%i] OK!!!\n", (*desc)->initial_user_rank);
			}

			free(payload);
		}
		else if(params.op_type == ACCUMULATE) {
			Acc_args args;

			memcpy(&args, target_ptr, sizeof(Acc_args));
			target_ptr += sizeof(Acc_args);

			assert(args.origin_count == node->origin_count);
			assert(args.origin_datatype == node->origin_datatype);
			assert(args.target_disp == node->target_disp);
			assert(args.target_count == node->target_count);
			assert(args.target_datatype == node->target_datatype);
			assert(args.operation == node->operation);

			int origin_type_size = -1;
			MPI_Type_size(args.origin_datatype, &origin_type_size);
			int payload_size = args.origin_count * origin_type_size;

			payload = (char*) malloc( payload_size );

			memcpy(payload, target_ptr, payload_size );
			target_ptr += payload_size;

			int j = 0;
			for(; j < payload_size; j++) {

				assert(payload[j] == node->payload[j]);
				//fprintf(stderr,"[%i] OK!!!\n", (*desc)->initial_user_rank);
			}

			free(payload);
		}   

		// someting bad happens 
		else {
			debug_message_gen_e( TRUE, stderr, " CM: ERROR: TEST LOGGING: sth bad happened, unknown operation type!!!\n");
		}
	}

	if(verbose == TRUE) {
		int left = MAX_LOCAL_LOG_SIZE - __testing_puts_nr;
		if(__testing_puts_nr != log_space->number_of_nodes_local) {
			debug_message_gen( TRUE, stderr, " CM: : TESTING: ERROR: __testing_puts_nr(%i) != log_space->number_of_nodes_local(%i)\n", __testing_puts_nr, log_space->number_of_nodes_local);
		}
		else {
			debug_message_gen( FALSE, stderr, " CM: : TESTING: OK: __testing_puts_nr(%i) == log_space->number_of_nodes_local(%i)\n", __testing_puts_nr, log_space->number_of_nodes_local);
		}
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: TESTING LOCAL LOGS finished CORRECTLY. Number of puts logged: %i, left: %i\n", __testing_puts_nr, left);
	}
}

void __delete_put_testing_logs(int rank_to_clear, int epoch_to_clear, int verbose) {
	Log_space* log_space = (*cm_desc)->log_space;
	if(verbose == TRUE) {
		int left = MAX_LOCAL_LOG_SIZE - __testing_puts_nr;
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: START DELETING LOCAL LOGS. Number of puts logged: %i, left: %i, log_space->number_of_nodes_local: %i\n",
											__testing_puts_nr, left, (*cm_desc)->log_space->number_of_nodes_local);
	}

	__Op_node** new_array = malloc( sizeof(__Op_node*) * MAX_LOCAL_LOG_SIZE );

	int i = 0;
	for(i = 0; i < MAX_LOCAL_LOG_SIZE; i++) new_array[i] = NULL;

	i = 0;
	for(; i < __testing_puts_nr; i++) {

		__Op_node* node = __local_log_array[i];

		if(node->target_rank == rank_to_clear && node->epoch_nr < epoch_to_clear) {

			free(node->payload);
			free(node);
			__local_log_array[i] = NULL;
		}
	}

	int j = 0;
	i = 0;
	for(; i < __testing_puts_nr; i++) {

		if(__local_log_array[i] != NULL) {

			new_array[j] = __local_log_array[i];
			j++;
		}
	}

	free(__local_log_array);
	__local_log_array = new_array;
	__testing_puts_nr = j;

	if(verbose == TRUE) {
		//	Log_space* log_space = (*cm_desc)->log_space;
		if(__testing_puts_nr != log_space->number_of_nodes_local) {
			debug_message_gen( TRUE, stderr, " CM: : TESTING DELETING: ERROR: __testing_puts_nr(%i) != log_space->number_of_nodes_local(%i)\n", __testing_puts_nr, log_space->number_of_nodes_local);
		}
		else {
			debug_message_gen( FALSE, stderr, " CM: : TESTING DELETING: OK: __testing_puts_nr(%i) == log_space->number_of_nodes_local(%i)\n", __testing_puts_nr, log_space->number_of_nodes_local);
		}

		int left = MAX_LOCAL_LOG_SIZE - __testing_puts_nr;
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: DELETING LOCAL LOGS finished CORRECTLY. Number of puts logged: %i, left: %i\n", __testing_puts_nr, left);
	}
}

void __delete_get_testing_logs(int rank_to_clear, int epoch_to_clear, int verbose) {
	Log_space* log_space = (*cm_desc)->log_space;
	if(verbose == TRUE) {
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: START DELETING LOCAL LOGS of gets for rank %i, epoch %i\n",
											rank_to_clear, epoch_to_clear);
	}

	__Op_node* current_elem = gdsl_queue_get_head(__global_log_array[rank_to_clear]);
	if(current_elem == NULL) {return;}
	//assert(current_elem != NULL);

	if(verbose == TRUE) {
		__Op_node* last_elem = gdsl_queue_get_tail(__global_log_array[rank_to_clear]);
		assert(last_elem != NULL);
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: first test GET node for rank %i has epoch %i, last: %i\n",
											rank_to_clear, current_elem->epoch_nr, last_elem->epoch_nr);
	}

	assert(current_elem->epoch_nr < epoch_to_clear);
	while(current_elem->epoch_nr < epoch_to_clear) {
		gdsl_queue_remove(__global_log_array[rank_to_clear]);
		current_elem = gdsl_queue_get_head(__global_log_array[rank_to_clear]);
		if(current_elem == NULL) {
			break;
		}
	}

	__testing_gets_nr[rank_to_clear] = gdsl_queue_get_size(__global_log_array[rank_to_clear]);

	if(verbose == TRUE) {
		if(gdsl_queue_get_size(__global_log_array[rank_to_clear]) != 0) {
		__Op_node* last_elem = gdsl_queue_get_tail(__global_log_array[rank_to_clear]);
		__Op_node* first_elem = gdsl_queue_get_head(__global_log_array[rank_to_clear]);
		debug_message_gen( FALSE, stderr, " CM: TEST LOGGING: LOCAL TEST GET LOGS DELETED for rank %i, first epoch: %i, last: %i\n", first_elem->epoch_nr, last_elem->epoch_nr);
		}
	}
}

//TODO: get logs to be freed
void __delete_all_testing_logs() {

	int i = 0;
	for(; i < __testing_puts_nr; i++) {

		free(__local_log_array[i]->payload);
		free(__local_log_array[i]);
	}

	i = 0;
	for(i = 0; i < MAX_LOCAL_LOG_SIZE; i++) __local_log_array[i] = NULL;

	__testing_puts_nr = 0;
}

void __delete_all_put_testing_logs() {
}

void __delete_all_get_testing_logs() {
}

void __finalize_logging_testing() {

	__delete_all_testing_logs();

	free(__local_log_array);
}

void __replay_actions() {
	replay_actions();
}
