#include "flib_win_free.h"

int fMPI_Win_free(foMPI_Win* win) {
#ifdef FLIB_DEBUG
	if((*desc)->flib_level_rank == 0) {
		fprintf(stderr,"[%i] Start win free...\n", (*desc)->initial_user_rank);
	}
#endif

	send_notification_to_assigned_cp_proc();

	MPI_Barrier((*desc)->initial_user_comm);

	int result = MPI_SUCCESS;

	if(cm_proc()) {

#ifdef FLIB_LOGGING_EXTENDED_DEBUG
		if(do_logging()) {
			__finalize_logging_testing();
		}
#endif

		/* finalize user application data */
		result = foMPI_Win_free(win);

		finalize_epoch_data();
		finalize_logging_data();
		finalize_cm_descriptor();
	}
	else {
		finalize_cp_descriptor();
	}

	finalize_checkpoint_data();
	finalize_groups();

	finalize_main_descriptor();

	return result;
}

void send_notification_to_assigned_cp_proc() {
	MPI_Request r;
	/* right before finalizing, CM processes send QUIT confirmation to the CP process*/
	MPI_Isend(NULL, 0, MPI_CHAR, my_assigned_cp_proc(), CM_QUIT, (*desc)->initial_user_comm, &r);
}

void finalize_epoch_data() {
	foMPI_Win_free((*cm_desc)->epoch_win);

	free((*cm_desc)->local_epoch_counter);
	free((*cm_desc)->epoch_win);
}

void finalize_logging_data() {
	if(!do_logging()) {
		return;
	}
	foMPI_Win_free((*cm_desc)->log_space->local_list);
	foMPI_Win_free((*cm_desc)->log_space->local_list_control);
	foMPI_Win_free((*cm_desc)->log_space->local_list_size);
	foMPI_Win_free((*cm_desc)->log_space->global_list);
	foMPI_Win_free((*cm_desc)->log_space->global_logs_control_win);
	foMPI_Win_free((*cm_desc)->log_space->puts_win);

	free((*cm_desc)->log_space->global_logs_control_win);
	free((*cm_desc)->log_space->local_sums);
	free((*cm_desc)->log_space->local_list);
	free((*cm_desc)->log_space->global_list);
	free((*cm_desc)->log_space->puts_win);	
	free((*cm_desc)->log_space->procs_communicated_using_gets);

	free((*cm_desc)->log_space); 
}

void finalize_cm_descriptor() {
	/*free((*cm_desc)->cp_procs_counts);

	int i = 0;
	for(; i < (*desc)->cp_size; i++) {
		free(((*cm_desc)->cp_procs_ranks_to_send)[i]);
	}
	free((*cm_desc)->cp_procs_ranks_to_send);*/

	foMPI_Win_free((*cm_desc)->marker_M_win);

	free((*cm_desc)->marker_M_space);
	free((*cm_desc)->sent_necessary_checkpoint_done);
	free((*cm_desc)->sent_checkpoint_request);
	free((*cm_desc)->sent_checkpoint_request_epoch);

	free(*cm_desc);
	free(cm_desc);
}

void finalize_checkpoint_data() {
	foMPI_Win_free((*desc)->cp_win);
	free((*desc)->cp_win);
}

void finalize_cp_descriptor() {
	if(do_logging()) {
		int i = 0;
		for(; i < (*desc)->initial_user_size; i++) {
			gdsl_queue_free((*cp_desc)->queued_demand_checkpoint_requests[i]);
		}
		free((*cp_desc)->queued_demand_checkpoint_requests);
	}

	free((*cp_desc)->received_coordinated_checkpoint);

	free(*cp_desc);
	free(cp_desc);
}

void finalize_main_descriptor() {
	free((*desc)->flib_parameters);
	free(*desc);
	free(desc);
}

void finalize_groups() {
	MPI_Group group;
	MPI_Comm_group(*(*desc)->flib_level_comm, &group);
	MPI_Group_free(&group);
}
