#include "flib.h"
#include "flib_cm_proc.h"
#include "flib_logging_protocol.h"
#include "flib_markers.h"

int fMPI_Put(const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype,
							foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Put(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, win);

	log_put(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype);

	(*desc)->remote_ops_done++;

	return result;
}

int fMPI_Get(void *origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype,
							foMPI_Win win) {
	cm_process_notifications();

	set_M_marker(target_rank);

	int result = foMPI_Get(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, win);

	log_get(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype);

	(*desc)->remote_ops_done++;

	return result;
}

int fMPI_Accumulate(const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, int target_rank, MPI_Aint target_disp, int target_count,
                     MPI_Datatype target_datatype, MPI_Op op, foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Accumulate(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, op, win);

	log_accumulate(origin_addr, origin_count, origin_datatype, target_rank, target_disp, target_count, target_datatype, op);

	(*desc)->remote_ops_done++;

	return result;
}

int fMPI_Get_accumulate(const void *origin_addr, int origin_count, MPI_Datatype origin_datatype, void *result_addr, int result_count, MPI_Datatype result_datatype,
                         int target_rank, MPI_Aint target_disp, int target_count, MPI_Datatype target_datatype, MPI_Op op, foMPI_Win win) {
	cm_process_notifications();

	set_M_marker(target_rank);

	int result = foMPI_Get_accumulate(origin_addr, origin_count, origin_datatype, result_addr, result_count, result_datatype,
																		target_rank, target_disp, target_count, target_datatype, op, win);

	log_get_accumulate(origin_addr, origin_count, origin_datatype, result_addr, result_count, result_datatype, target_rank, target_disp, target_count, target_datatype, op);

	(*desc)->remote_ops_done++;

	return result;
}

int fMPI_Fetch_and_op(const void *origin_addr, void *result_addr, MPI_Datatype datatype, int target_rank, MPI_Aint target_disp, MPI_Op op, foMPI_Win win) {
	cm_process_notifications();

	set_M_marker(target_rank);

	int result = foMPI_Fetch_and_op(origin_addr, result_addr, datatype, target_rank, target_disp, op, win);

	log_fetch_and_op(origin_addr, result_addr, datatype, target_rank, target_disp, op);

	(*desc)->remote_ops_done++;

	return result;
}

int fMPI_Compare_and_swap(const void *origin_addr, const void *compare_addr, void *result_addr, MPI_Datatype datatype, int target_rank, MPI_Aint target_disp, foMPI_Win win) {
	cm_process_notifications();

	set_M_marker(target_rank);

	int result = foMPI_Compare_and_swap(origin_addr, compare_addr, result_addr, datatype, target_rank, target_disp, win);

	log_compare_and_swap(origin_addr, compare_addr, result_addr, datatype, target_rank, target_disp);

	(*desc)->remote_ops_done++;

	return result;
}


