#include "flib.h"
#include "flib_logging_protocol.h"
#include "flib_cm_proc.h"

int fMPI_Win_fence(int assert, foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Win_fence(assert, win);

	//TODO: change this a little unfortunate name
  (*cm_desc)->superstep++;

	cm_manage_coordinated_checkpoint();
	//cm_manage_global_recovery();
	log_all_outstanding_operations();
	inc_local_epoch_counters();
	//inc_global_epoch_counters();

	inc_GNC_cnt();

	return result;
}

int fMPI_Win_lock(int lock_type, int rank, int assert, foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Win_lock(lock_type, rank, assert, win);

	inc_local_epoch_counter(rank);
	inc_global_epoch_counter(rank);

	return result;
}

int fMPI_Win_unlock(int rank, foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Win_unlock(rank, win);

	log_outstanding_operations(rank);

	inc_local_epoch_counter(rank);
	inc_global_epoch_counter(rank);

	return result;
}

int fMPI_Win_lock_all(int assert, foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Win_lock_all(assert, win);

	inc_local_epoch_counters();
	inc_global_epoch_counters();

	return result;
}

int fMPI_Win_unlock_all(foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Win_unlock_all(win);

	log_all_outstanding_operations();

	inc_local_epoch_counters();
	inc_global_epoch_counters();

	return result;
}

int fMPI_Win_flush(int rank, foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Win_flush(rank, win);

	log_outstanding_operations(rank);

 	inc_local_epoch_counter(rank);
 // inc_global_epoch_counter(rank);

	return result;
}

int fMPI_Win_flush_all(foMPI_Win win) {
	cm_process_notifications();
	
	int result = foMPI_Win_flush_all(win);

	log_all_outstanding_operations();

  inc_local_epoch_counters();
  inc_global_epoch_counters();

	return result;
}

//TODO: for now the same as usual flush
int fMPI_Win_flush_local(int rank, foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Win_flush_local(rank, win);

	log_outstanding_operations(rank);

  inc_local_epoch_counter(rank);
  inc_global_epoch_counter(rank);

	return result; 
}

//TODO: as above
int fMPI_Win_flush_local_all(foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Win_flush_local_all(win);

	log_all_outstanding_operations();
  
	inc_local_epoch_counters();
  inc_global_epoch_counters();

  return result;
}

//TODO: check this thing. 
int fMPI_Win_sync(foMPI_Win win) {
	cm_process_notifications();

	int result = foMPI_Win_sync(win);

  inc_local_epoch_counters();

	return result;
}

