/*
 * Copyright (c) 2012 Torsten Hoefler. All rights reserved.
 *
 * Author(s): Torsten Hoefler <htor@illinois.edu>
 *
 */

#include <stdio.h>
#include <stdlib.h>
//#include "../MersenneTwister.h"
//#include "common.h"
#include <time.h>

#define SIZE 8

#define KUKU RAND_MAX*RAND_MAX

typedef struct{
  int value;
  int next;
} t_elem;

typedef struct {
  t_elem *table; // the table + heap
  int *last; // direct link to last element -- avoid traversing list!
  int nextfree; // the next free item in heap
  int size; // size of the heap
  int tsize; // size of table
} t_hash;

int hashfunc(int val) {
  return val & (SIZE-1);
}

void insert(t_hash *hash, int elem) {

  //printf(">>>>>>>> insert %i\n", elem);
  int pos = hashfunc(elem);
  printf("######## insert %i at position %i\n", elem,pos);
  if(hash->table[pos].value == -1) {
    printf(">>> normal insert...\n");
    hash->table[pos].value = elem;
  } else {
    printf(">>> collision: h->table[pos].value = %i, elem = %i, pos = (%i)\n", hash->table[pos].value, elem, hashfunc(elem));
    // check heap and go if necessary
    int newelem=hash->nextfree++;
    //int next = hash->nextfree;
    //while(hash->table[next] != -1) {
    //  next++;
    //}
    

    //printf(">>> h->nextfree (newelem) = %i\n",newelem);
    //printf("allocated %i\n", hash->nextfree-SIZE);
    if(hash->table[pos].next == -1) {
      //printf(">>> h->table[pos].next == -1; new value = %i (newelem)\n",newelem);
      // link new elem from table
      hash->table[pos].next = newelem;
    } else {
      int newpos=hash->last[pos];
     hash->table[newpos].next = newelem;

     //printf(">>> h->table[pos].next == %i; newpos = %i\n",hash->table[pos].next,newpos);
    }
    hash->last[pos]=newelem;
    // fill allocated element
    hash->table[newelem].value = elem;

    
  }
}

int deleteElem(t_hash* hash, int pos) {
  printf("$$$$ delete (pos: %i)\n", pos);

  if(hash->last[pos] == -1) {
    hash->table[pos].value = -1;
  }
  else {
    int offset = hash->last[pos];
    int value = hash->table[offset].value;
    //hash->table[offset].value = -1;
    
    int prevOffset = pos;
    int prevValue = hash->table[prevOffset].value;
    
    int nextOffset = hash->table[prevOffset].next;
    int nextValue = hash->table[nextOffset].value;

    while(1) {
      nextOffset = hash->table[prevOffset].next;
      nextValue = hash->table[nextOffset].value;
      
      if(nextValue == -1) {
	return -1;
      }

      if(nextOffset == offset) {
	hash->table[prevOffset].next = -1;
	hash->last[pos] = prevOffset;
	hash->table[offset].value = -1;
	
	return nextValue;
      }

      prevOffset = hash->table[prevOffset].next;
      prevValue = hash->table[prevOffset].value;
    }
  }
}


int count(t_hash *hash) {

  int cnt=0;
  int heapcnt=0;
  for(int i=0; i<hash->tsize; ++i) {
    int pos = i;
    if(hash->table[pos].value != -1) {
      cnt++;
      if(hash->table[pos].next != -1) {
        heapcnt++;
        pos=hash->table[pos].next;
        while(hash->table[pos].next != -1) {
          pos = hash->table[pos].next;
          heapcnt++;
        }
      }
    }
  }
  printf("cnt: %i heapcnt: %i, total: %i\n", cnt, heapcnt, cnt+heapcnt);
  return cnt+heapcnt;
}

void printTable(t_hash* h) {
  int i = 0;
  printf("@@@@@@ Table:\n");
  for(i = 0; i < h->tsize; i++) {
    printf("%i (value: %i, next: %i)\n",i, h->table[i].value, h->table[i].next);
  }
  printf("@@@ Heap:\n");
  for(i = h->tsize; i < (h->tsize+h->size); i++) {
    printf("%i (value: %i, next: %i)\n",i, h->table[i].value, h->table[i].next);
  }
  printf("@@@ Last:\n");
  for(i = 0; i < h->tsize; i++) {
    printf("%i (value: %i)\n",i, h->last[i]);
  }
  printf("@@@ Nextfree: %i\n\n",h->nextfree);
  // printf("\n");
}

int main(int argc, char **argv) {

  //  int n = atoi(argv[1]); // number of total entries, will fill 1/2

  srand(time(NULL));
  printf("RAND MAX: %i\n",RAND_MAX);
  printf("KUKU: %llu\n",(int64_t)KUKU);

  for(int i = 0; i < 10; i++) {
    printf("%i\n",rand());
    printf("%i\n",rand() & 65535);
  }

  printf("--\n");
  int64_t val;

  for(int i = 0; i < 10; i++) {
    val = rand();
    val = val << 32;
    //printf("%llu\n",bigRandVal());
    //printf("%llu\n",bigRandVal() & 65535);
  }

  t_hash hash;

  /*  hash.tsize = SIZE;
  hash.table = (t_elem*)calloc(1,(hash.tsize+n)*sizeof(t_elem)); // the actual table
  hash.last = (int*)calloc(1,hash.tsize*sizeof(int));
  hash.nextfree = hash.tsize;
  hash.size = n;
  
  MTRand mt(999);
  
  // initialize table
   for(int i=0; i<hash.tsize+n; ++i) {
    if(i<hash.tsize) hash.last[i]=-1;
    hash.table[i].next=-1;
    hash.table[i].value=-1;
  }

  for(int i=0; i<n; ++i) {
    printTable(&hash);
    insert(&hash, abs(mt.randInt())); // positive
    printTable(&hash);
    printf("\n");
  }

  deleteElem(&hash,0);
  printTable(&hash);
  deleteElem(&hash,0);
  printTable(&hash);
  
  deleteElem(&hash,2);
  printTable(&hash);

  deleteElem(&hash,4);
  printTable(&hash);

  deleteElem(&hash,0);
  printTable(&hash);

  deleteElem(&hash,1);
  printTable(&hash);

  deleteElem(&hash,3);
  printTable(&hash);

  deleteElem(&hash,5);
  printTable(&hash);
  
  deleteElem(&hash,5);
  printTable(&hash);

  count(&hash);
  */ 
}
