#include <mpi.h>
#include <stdlib.h>
#include "fompi.h"
#include "flib.h"
#include "flib_tests.h"

#include <math.h>
#include <assert.h>

//#define MAX_RANDOM 100
//#define MAX_COUNT 10
//#define MAX_SIZE 100
#define MTBF -1000

void local_computation();

int main( int argc, char *argv[] ) {

	/* ================== flib-connected data structures */

	/* communicator that enables to distinguish between compute (CM) and checkpoint (CP) processes */
	/* user has to pass this communicator to the modified flib-specific win create function */
	/* it has to use this communicator instead of their original one during the computation */
	MPI_Comm flib_level_comm;

	int user_rank = -1;
	int user_size = -1;

	/* ================== end of flib-connected data structures */


	/* we assume that the application uses one window */
	foMPI_Win app_win;

	//char* base; 
	double* base = NULL;
	//uint64_t* base;

	int world_size;
	int world_rank;

	foMPI_Init( &argc, &argv );

	MPI_Comm_size( MPI_COMM_WORLD, &world_size );
	MPI_Comm_rank( MPI_COMM_WORLD, &world_rank );

	// ================== FLIB code: determining the number of CP devices
	// we assume for now that user simply passes the number of CP devices as the input argument
	// user can also pass e.g. a ratio/percentage as the input argument  

	int cp_procs = atoi(argv[1]); 
	int win_size = atoi(argv[2]);
	int log_size = atoi(argv[3]);;
	int ITERS = atoi(argv[4]);
	int MAX_RANDOM = atoi(argv[5]);
	int MAX_COUNT = atoi(argv[6]);
	int MAX_SIZE = atoi(argv[7]);
	int log_puts = FALSE;
	int log_gets = TRUE;
	int protocol_type = FLIB_NO_COORDINATED_SCHEME;
	int hierarchy = _FLIB_NONE_HIERARCHY;

	if(world_rank == 0) {
		printf("CP processes: %i\n", cp_procs);
	}

	MPI_Aint app_win_size = win_size*sizeof(double);
	uint32_t app_win_disp_unit = sizeof(double);	

	//MPI_Aint app_win_size = 2*sizeof(char);
	//uint32_t app_win_disp_unit = sizeof(char);

	//MPI_Aint app_win_size = 2*sizeof(uint64_t); 
	//uint32_t app_win_disp_unit = sizeof(uint64_t);

	base = (double*) malloc( win_size*sizeof(double) ); 
	//base = (uint64_t*) malloc( 2*sizeof(uint64_t) );
	//base = (char*) malloc( 2*sizeof(char) );    

	for(int i = 0; i < win_size; i++) {
		base[i] = (double)(rand() % MAX_RANDOM) / (double)((rand() + 1) % MAX_RANDOM);
	}

	/* IMPORTANT: we have to allocate the memory for the user window BEFORE calling fMPI_Win_create */

	/* this function inits all the FLIB data structures, regroups the processes - divides them into CM's and CP's */
	/* only CM's return from this function */
	fMPI_Win_create( base, app_win_size, app_win_disp_unit, MPI_INFO_NULL, MPI_COMM_WORLD, &app_win, cp_procs, &flib_level_comm, hierarchy, MTBF, protocol_type, log_puts, log_gets, log_size );

	//the user code
	//communication using flib_level_comm

	//the user rank
	//rank in the group of the computing processes
	user_rank = fLIB_get_user_rank();

	//the number of computing processes
	user_size = fLIB_get_user_size();

	//the user code
	//I use a very simple BSP model of computation for testing purposes
	//communication using flib_level_comm

	srand(10000000*user_rank);

	//a necessary addition to the user code
	//the user code has to be included in such
	//a region
	//the region has to end right before calling
	//fMPI_Win_free

	if(cm_proc()) {
		int i = 1;

		int* var = (int*) malloc(sizeof(int));
		for(; i < ITERS; i++) {		

#ifdef FLIB_DEBUG
			if((*desc)->flib_level_rank == 0) {
				fprintf(stderr,"[%i] beginning stage %i of the computation\n", (*desc)->initial_user_rank, i);
			}
#endif

			//some fake local time-consuming computation
			local_computation();

			int target_rank = rand() % user_size;

#ifdef FLIB_VERBOSE_2_DEBUG
			fprintf(stderr,"[%i] ^^^ before gets %i\n", (*desc)->initial_user_rank, target_rank);
#endif

			//foMPI_Accumulate( &base[0], 1, MPI_UINT64_T, target_rank, 1, 1, MPI_UINT64_T, foMPI_SUM, app_win );
			//foMPI_Put( &base[0], 1, MPI_UINT64_T, target_rank, 1, 1, MPI_UINT64_T, app_win );
			//fMPI_Put( &base[0], 1, MPI_DOUBLE, target_rank, 1, 1, MPI_DOUBLE, app_win );
			int count = rand() % MAX_COUNT;

			for(int j = 0; j < count; j++) {
				int disp = rand() % win_size;
				int size = rand() % MAX_SIZE;

				while(disp + size >= win_size) {
					disp = rand() % win_size;
					size = rand() % MAX_SIZE;
				}

				int* my_var = (int*) malloc(size * sizeof(int));

				fMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, target_rank, 0, app_win);

				fMPI_Get( my_var, 1, MPI_DOUBLE, target_rank, disp, 1, MPI_DOUBLE, app_win );
				fMPI_Win_flush(target_rank, app_win);

				fMPI_Win_unlock(target_rank, app_win);
				
				free(my_var);	
			}

#ifdef FLIB_VERBOSE_2_DEBUG
			fprintf(stderr,"[%i] ^^^ after gets in %i\n", (*desc)->initial_user_rank, target_rank);
#endif

			//fMPI_Win_flush(target_rank, app_win);	

			//foMPI_Win_sync( app_win );

			__test_all_logs_of_gets(TRUE);

			fMPI_Win_fence( MPI_INFO_NULL, app_win );
		}
		free(var);
	}

	fMPI_Win_free( &app_win );

	foMPI_Finalize();

	return 0;
}

//simple local computation
void local_computation() {
	double w = 1;
	int k = 0;
	for(; k < 50000000; k++) {
		w = w / rand() + sin(k) * cos(k);
	}

}

