#include <mpi.h>
#include <stdlib.h>
#include "fompi.h"
#include "flib.h"

#include <math.h>
#include <assert.h>


void local_computation();

int main( int argc, char *argv[] ) {

	/* ================== flib-connected data structures */

	/* communicator that enables to distinguish between compute (CM) and checkpoint (CP) processes */
	/* user has to pass this communicator to the modified flib-specific win create function */
	/* it has to use this communicator instead of their original one during the computation */
	MPI_Comm flib_level_comm;

	int user_rank = -1;
	int user_size = -1;

	/* ================== end of flib-connected data structures */

	
	/* we assume that the application uses one window */
  foMPI_Win app_win;
  
	//char* base; 
	double* base = NULL;
	//uint64_t* base;

  int world_size;
  int world_rank;

  foMPI_Init( &argc, &argv );

  MPI_Comm_size( MPI_COMM_WORLD, &world_size );
  MPI_Comm_rank( MPI_COMM_WORLD, &world_rank );

	// ================== FLIB code: determining the number of CP devices
	// we assume for now that user simply passes the number of CP devices as the input argument
	// user can also pass e.g. a ratio/percentage as the input argument  

	int cp_procs = atoi(argv[1]); 

	if(world_rank == 0) {
		printf("CP processes: %i\n", cp_procs);
	}
	

	MPI_Aint app_win_size = 2*sizeof(double);
	uint32_t app_win_disp_unit = sizeof(double);	

	//MPI_Aint app_win_size = 2*sizeof(char);
	//uint32_t app_win_disp_unit = sizeof(char);

	//MPI_Aint app_win_size = 2*sizeof(uint64_t); 
	//uint32_t app_win_disp_unit = sizeof(uint64_t);

  base = (double*) malloc( 2*sizeof(double) ); 
	//base = (uint64_t*) malloc( 2*sizeof(uint64_t) );
	//base = (char*) malloc( 2*sizeof(char) );    

	base[0] = user_rank + 50.23;
  base[1] = -40.11 / (1 + 10000*user_rank);

	/* IMPORTANT: we have to allocate the memory for the user window BEFORE calling fMPI_Win_create */

	/* this function inits all the FLIB data structures, regroups the processes - divides them into CM's and CP's */
	/* only CM's return from this function */
	fMPI_Win_create( base, app_win_size, app_win_disp_unit, MPI_INFO_NULL, MPI_COMM_WORLD, &app_win, cp_procs, &flib_level_comm );

//the user code
  //communication using flib_level_comm

  //the user rank
  //rank in the group of the computing processes
  user_rank = get_user_rank();

  //the number of computing processes
  user_size = get_user_size();

#ifdef FLIB_VERBOSE_DEBUG

	fprintf(stderr,"[initial: %i, flib: %i] computing: %i \n", (*desc)->initial_user_rank, (*desc)->flib_level_rank, (*desc)->compute_proc);

#endif

	//the user code
	//I use a very simple BSP model of computation for testing purposes
	//communication using flib_level_comm

	/*TODO: discuss this with Torsten*/

	srand(10000000*user_rank);

	if(cm_proc()) {

		int i = 1;
		for(; i < 20; i++) {		

#ifdef FLIB_DEBUG

				if((*desc)->flib_level_rank == 0) {
					fprintf(stderr,"[%i] beginning stage %i of the computation\n", (*desc)->initial_user_rank, i);
				}

#endif

			//some fake local time-consuming computation

			//base[0] = 8888 + i;		
			local_computation();
			
			//base[0] = base[0] * ((double) rand() % 50) / ((double) rand() % 50) * ( (rand() % 2 == 0) ? -1 : 1);

			int target_rank = rand() % user_size;

			while(target_rank == user_rank) {
		
				target_rank = rand() % user_size;
			}

			// TODO: IMPORTANT!!!!!!!!!!!!!!: what if the user app uses self locks? danger!!!!!

			//int target_rank = (user_rank + 1) % user_size;
	
#ifdef FLIB_VERBOSE_2_DEBUG	
			fprintf(stderr,"[%i] ^^^ before locking %i\n", (*desc)->initial_user_rank, target_rank);
#endif	

			fMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, target_rank, 0, app_win);

#ifdef FLIB_VERBOSE_2_DEBUG
			fprintf(stderr,"[%i] ^^^ after locking %i\n", (*desc)->initial_user_rank, target_rank);
#endif

	  	//foMPI_Accumulate( &base[0], 1, MPI_UINT64_T, target_rank, 1, 1, MPI_UINT64_T, foMPI_SUM, app_win );
			//foMPI_Put( &base[0], 1, MPI_UINT64_T, target_rank, 1, 1, MPI_UINT64_T, app_win );
	
			fMPI_Put( &base[0], 1, MPI_DOUBLE, target_rank, 1, 1, MPI_DOUBLE, app_win );

#ifdef FLIB_VERBOSE_2_DEBUG
			fprintf(stderr,"[%i] ^^^ after put in %i\n", (*desc)->initial_user_rank, target_rank);
#endif

			fMPI_Win_flush(target_rank, app_win);	
	
#ifdef FLIB_VERBOSE_2_DEBUG
			fprintf(stderr,"[%i] ^^^ after flushing %i\n", (*desc)->initial_user_rank, target_rank);
#endif

			fMPI_Win_unlock(target_rank, app_win);	

#ifdef FLIB_VERBOSE_2_DEBUG
			fprintf(stderr,"[%i] ^^^ after unlocking %i\n", (*desc)->initial_user_rank, target_rank);
#endif

			//foMPI_Win_sync( app_win );

#ifdef FLIB_DEBUG

    	// ================== FLIB code
    	//before I finish logging protocol, in this section I simulate adaptive checkpointing

    	//if(user_rank == 1 && i == 2) {  
    	if((rand() + (*desc)->initial_user_rank) % 15 == 0) {
    		__simulate_adaptive_checkpoint();
    	}

#endif

	  	fMPI_Win_fence( MPI_INFO_NULL, app_win );
		}
	
#ifdef FLIB_DEBUG

	MPI_Barrier(flib_level_comm);
	
	__test_local_logs();

#endif

	}

	/* TODO: we need to somehow make this library SAFE !!! */
	/* for example, right now, if the user places here (right before win free 
		 a barrier with mpi comm world, then we have a deadlock.
		 The solution is to demand placing win free RIGHT after the CM block */

  fMPI_Win_free( &app_win );

	MPI_Barrier(MPI_COMM_WORLD);
	
  foMPI_Finalize();

  return 0;
}

//simple local computation
void local_computation() {

	double w = 1;
	int k = 0;
	for(; k < 50000000; k++) {
		w = w / rand() + sin(k) * cos(k);
	}

}

