#include <mpi.h>
#include <stdlib.h>
#include "fompi.h"
#include "flib.h"
#include "flib_tests.h"
#include "flib_debug.h"

#include <math.h>
#include <assert.h>

//#define MAX_RANDOM 100
//#define MAX_COUNT 10
//#define MAX_SIZE 100
#define MTBF -1000

int main( int argc, char *argv[] ) {

	/* ================== flib-connected data structures */

	/* communicator that enables to distinguish between compute (CM) and checkpoint (CP) processes */
	/* user has to pass this communicator to the modified flib-specific win create function */
	/* it has to use this communicator instead of their original one during the computation */
	MPI_Comm flib_level_comm;

	int user_rank = -1;
	int user_size = -1;

	/* ================== end of flib-connected data structures */


	/* we assume that the application uses one window */
	foMPI_Win app_win;

	uint64_t* base;

	int world_size;
	int world_rank;

	foMPI_Init( &argc, &argv );

	MPI_Comm_size( MPI_COMM_WORLD, &world_size );
	MPI_Comm_rank( MPI_COMM_WORLD, &world_rank );

	// ================== FLIB code: determining the number of CP devices
	// we assume for now that user simply passes the number of CP devices as the input argument
	// user can also pass e.g. a ratio/percentage as the input argument  

	int cp_procs = atoi(argv[1]); 
	int win_size = atoi(argv[2]);
	int log_size = atoi(argv[3]);;
	int ITERS = atoi(argv[4]);
	int MAX_RANDOM = atoi(argv[5]);
	int MAX_COUNT = atoi(argv[6]);
	int MAX_SIZE = atoi(argv[7]);
	int log_puts = atoi(argv[8]);
	int log_gets = atoi(argv[9]);
	int protocol_type = atoi(argv[10]);
	int hierarchy = _FLIB_NONE_HIERARCHY;

	if(world_rank == 0) {
		printf("CP processes: %i\n", cp_procs);
	}

	MPI_Aint app_win_size = world_size*sizeof(uint64_t);
	uint32_t app_win_disp_unit = sizeof(uint64_t);	

	base = (uint64_t*) malloc( win_size*sizeof(uint64_t) );

	for(int i = 0; i < win_size; i++) {
		base[i] = rand() % MAX_RANDOM;
	}

	/* IMPORTANT: we have to allocate the memory for the user window BEFORE calling fMPI_Win_create */

	/* this function inits all the FLIB data structures, regroups the processes - divides them into CM's and CP's */
	/* only CM's return from this function */
	//fMPI_Win_create( base, app_win_size, app_win_disp_unit, MPI_INFO_NULL, MPI_COMM_WORLD, &app_win, cp_procs, &flib_level_comm, hierarchy, MTBF, protocol_type, log_puts, log_gets, log_size );

	fMPI_Win_create( base, app_win_size, app_win_disp_unit, MPI_INFO_NULL, MPI_COMM_WORLD, &app_win, cp_procs, &flib_level_comm, hierarchy, MTBF, protocol_type, log_puts, log_gets, log_size );

	//the user code
	//communication using flib_level_comm

	//the user rank
	//rank in the group of the computing processes
	user_rank = fLIB_get_user_rank();

	//the number of computing processes
	user_size = fLIB_get_user_size();

	//the user code
	//I use a very simple BSP model of computation for testing purposes
	//communication using flib_level_comm

	//a necessary addition to the user code
	//the user code has to be included in such a region
	//the region has to end right before calling fMPI_Win_free

	if(cm_proc()) {
		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "Begin phase 1 of tests\n");

		fLIB_take_ordered_coordinated_checkpoint();

		int target_rank = 0;

		uint64_t value = user_rank * 10 + 1;
		fMPI_Put(&value, 1, MPI_UINT64_T, target_rank, user_rank, 1, MPI_UINT64_T, app_win);
		fMPI_Win_flush(target_rank, app_win);

		MPI_Barrier(flib_level_comm);

		if(user_rank == target_rank) {
			for(int i = 0; i < user_size; i++) {
				fprintf(stderr,"%llu ", base[i]);
			}
			fprintf(stderr,"\n");
		}

		MPI_Barrier(flib_level_comm);
	
		if(user_rank == target_rank) {
			memset(&base[0], 0, app_win_size);

			__replay_actions();
		}
		
		MPI_Barrier(flib_level_comm);
		
		if(user_rank == target_rank) {
			for(int i = 0; i < user_size; i++) {
				fprintf(stderr,"%llu ", base[i]);
			}
			fprintf(stderr,"\n");
		}

		if(user_rank == target_rank) {
			for(int i = 0; i < user_size; i++) {
				if(i == (*desc)->flib_level_rank) {
					continue;
				}
				assert(base[i] == i * 10 + 1);
			}
		}

		MPI_Barrier(flib_level_comm);
		
		debug_message_e( _FLIB_FIRST_IN_FLIB_CM, FALSE, stderr, "Phase 1 completed successfully\n");
	}

	fMPI_Win_free( &app_win );

	foMPI_Finalize();

	return 0;
}

